from typing import Any, Optional

from buildingspy.io.outputfile import Reader  # type: ignore
from pydantic import BaseModel, ConfigDict, Field, model_validator
from yattag import Doc, SimpleDoc

from trano.elements import Space
from trano.plot.plot import plot_plot_ly_many
from trano.reporting.reporting import BaseDocumentation, ModelDocumentation


class HtmlDoc(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    doc: SimpleDoc = Field(default=None)  # type: ignore
    tag: Any = Field(default=None)
    text: Any = Field(default=None)

    @model_validator(mode="after")
    def instantiate_from_package(self) -> "HtmlDoc":
        doc, tag, text = Doc().tagtext()
        self.doc = self.doc or doc
        self.tag = self.tag or tag
        self.text = self.text or text
        return self


def to_html(
    documentation: BaseDocumentation, html_doc: Optional[HtmlDoc] = None
) -> str:
    html_doc = html_doc or HtmlDoc()
    if not documentation.table:
        return ""
    with html_doc.tag("body"):
        with html_doc.tag("h1", style="font-size: 16px;"):
            html_doc.text(documentation.title or documentation.topic)

        with html_doc.tag("p"), html_doc.tag("p"):
            if documentation.introduction:
                html_doc.text(documentation.introduction)
        with html_doc.tag("p"):
            if isinstance(documentation.table, list):
                for t in documentation.table:
                    html_doc.doc.asis(t.to_html())
            else:
                html_doc.doc.asis(documentation.table.to_html())
        with html_doc.tag("p"), html_doc.tag("p"):
            if documentation.conclusions:
                html_doc.text(documentation.conclusions)
    return html_doc.doc.getvalue()


def to_html_reporting(documentation: ModelDocumentation) -> str:
    html_doc = HtmlDoc()
    with html_doc.tag("html"):
        with html_doc.tag("head"), html_doc.tag("title"):
            if documentation.title:
                html_doc.text(documentation.title)
        html_doc.doc.asis(to_html(documentation.spaces))
        html_doc.doc.asis(to_html(documentation.constructions))
        html_doc.doc.asis(to_html(documentation.systems))
        space = [s for s in documentation.elements if isinstance(s, Space)]
        if documentation.result:
            mat = Reader(documentation.result.path, documentation.result.type)
            for figures in list(zip(*[s.figures for s in space])):
                figures_plotly = plot_plot_ly_many(mat, list(figures), show=False)
                if figures_plotly:
                    html_doc.doc.asis(figures_plotly.to_html())

    return html_doc.doc.getvalue()



from math import ceil
from typing import ClassVar, List, Optional, Union, TYPE_CHECKING

from networkx import Graph
from pydantic import Field, BaseModel


from trano.elements.base import BaseElement
from trano.elements.envelope import (
    BaseExternalWall,
    BaseFloorOnGround,
    BaseInternalElement,
    BaseWindow,
    ExternalWall,
    FloorOnGround,
    InternalElement,
    MergedBaseWall,
    MergedExternalWall,
    MergedWindows,
    WallParameters,
    WindowedWallParameters,
    VerticalWallParameters,
    RoofWallParameters,
    ExternalWallParameters,
)
from trano.elements.system import BaseOccupancy, Emission, System, AirHandlingUnit
from trano.elements.types import ContainerTypes

if TYPE_CHECKING:
    from trano.topology import Network

MAX_X_SPACES = 3


def _get_controllable_element(elements: List[System]) -> Optional["System"]:
    controllable_elements = []
    for element in elements:
        controllable_ports = element.get_controllable_ports()
        if controllable_ports:
            controllable_elements.append(element)
    if len(controllable_elements) > 1:
        raise NotImplementedError
    if not controllable_elements:
        return None
    return controllable_elements[0]


class BoundaryParameter(BaseModel):
    number_orientations: int = 0
    area_per_orientation: List[float] = Field(default_factory=lambda: [0.0])
    average_resistance_external: float = Field(0.001)
    average_resistance_external_remaining: float = Field(0.001)
    total_thermal_capacitance: float = Field(10000)
    tilts: List[float] = Field(default_factory=lambda: [0.0])
    azimuths: List[float] = Field(default_factory=lambda: [0.0])
    average_u_value: float = 0.001
    total_thermal_resistance: float = 0.001

    @classmethod
    def from_parameter(cls, parameter: WallParameters) -> "BoundaryParameter":
        if not parameter.number and parameter.type != "BaseWindow":
            return cls()
        return cls(
            number_orientations=parameter.number,
            area_per_orientation=(
                parameter.window_area_by_orientation
                if parameter.type == "BaseWindow"
                else parameter.surfaces
            ),
            average_resistance_external=parameter.average_resistance_external,
            average_resistance_external_remaining=parameter.average_resistance_external_remaining,
            total_thermal_capacitance=parameter.total_thermal_capacitance,
            tilts=parameter.tilts_to_radians(),
            azimuths=parameter.azimuths_to_radians(),
            average_u_value=parameter.average_u_value,
            tottal_thermal_resistance=parameter.total_thermal_resistance,
        )


class BoundaryParameters(BaseModel):
    roofs: BoundaryParameter = Field(default_factory=BoundaryParameter)
    external_boundaries: BoundaryParameter = Field(default_factory=BoundaryParameter)
    vertical_walls: BoundaryParameter = Field(default_factory=BoundaryParameter)
    windows: BoundaryParameter = Field(default_factory=BoundaryParameter)
    floors: BoundaryParameter = Field(default_factory=BoundaryParameter)

    @classmethod
    def from_boundaries(cls, parameters: List[WallParameters]) -> "BoundaryParameters":
        data = {}
        for p in parameters:
            if p.type == "ExternalWallRoof":
                data["roofs"] = BoundaryParameter.from_parameter(p)
            elif p.type == "ExternalWallVerticalOnly":
                data["vertical_walls"] = BoundaryParameter.from_parameter(p)
            elif p.type == "ExternalWallExternal":
                data["external_boundaries"] = BoundaryParameter.from_parameter(p)
            elif p.type == "FloorOnGround":
                data["floors"] = BoundaryParameter.from_parameter(p)
            elif p.type == "BaseWindow":
                data["windows"] = BoundaryParameter.from_parameter(p)
        return cls(**data)


class BaseSpace(BaseElement):
    counter: ClassVar[int] = 0
    name: str
    external_boundaries: list[
        Union["BaseExternalWall", "BaseWindow", "BaseFloorOnGround"]
    ]
    internal_elements: List["BaseInternalElement"] = Field(default=[])
    boundaries: Optional[List[WallParameters]] = None
    emissions: List[System] = Field(default=[])
    ventilation_inlets: List[System] = Field(default=[])
    ventilation_outlets: List[System] = Field(default=[])
    occupancy: Optional[BaseOccupancy] = None
    container_type: ContainerTypes = "envelope"
    boundary_parameters: Optional[BoundaryParameters] = None

    def model_post_init(self, __context) -> None:  # type: ignore # noqa: ANN001
        self._assign_space()

    def _assign_space(self) -> None:
        for emission in (
            self.emissions + self.ventilation_inlets + self.ventilation_outlets
        ):
            if emission.control:
                emission.control.space_name = self.name
        if self.occupancy:
            self.occupancy.space_name = self.name

    @property
    def number_merged_external_boundaries(self) -> int:
        return sum(
            [
                boundary.length
                for boundary in self.merged_external_boundaries + self.internal_elements
            ]
        )

    @property
    def number_ventilation_ports(self) -> int:
        return 2 + 1  # databus

    @property
    def merged_external_boundaries(
        self,
    ) -> List[
        Union["BaseExternalWall", "BaseWindow", "BaseFloorOnGround", "MergedBaseWall"]
    ]:

        external_walls = [
            boundary
            for boundary in self.external_boundaries
            if boundary.type in ["ExternalWall", "ExternalDoor"]
        ]
        windows = [
            boundary
            for boundary in self.external_boundaries
            if boundary.type == "Window"
        ]
        merged_external_walls = MergedExternalWall.from_base_elements(external_walls)
        merged_windows = MergedWindows.from_base_windows(windows)  # type: ignore
        external_boundaries: list[
            BaseExternalWall | BaseWindow | BaseFloorOnGround | MergedBaseWall
        ] = (
            merged_external_walls
            + merged_windows
            + [
                boundary
                for boundary in self.external_boundaries
                if boundary.type not in ["ExternalWall", "Window", "ExternalDoor"]
            ]
        )
        return external_boundaries

    def get_controllable_emission(self) -> Optional["System"]:
        return _get_controllable_element(self.emissions)

    def assign_position(self) -> None:
        x, y = [
            250 * (Space.counter % MAX_X_SPACES),
            150 * ceil(Space.counter / MAX_X_SPACES),
        ]
        self.position.set_global(x, y)
        Space.counter += 1

        for i, emission in enumerate(self.emissions):
            emission.position.set_global(x + i * 30, y - 75)
        if self.occupancy:
            self.occupancy.position.set_global(x - 50, y)

    def find_emission(self) -> Optional["Emission"]:
        emissions = [
            emission for emission in self.emissions if isinstance(emission, Emission)
        ]
        if not emissions:
            return None
        if len(emissions) != 1:
            raise NotImplementedError
        return emissions[0]

    def first_emission(self) -> Optional["System"]:
        if self.emissions:
            return self.emissions[0]
        return None

    def last_emission(self) -> Optional["System"]:
        if self.emissions:
            return self.emissions[-1]
        return None

    def get_ventilation_inlet(self) -> Optional["System"]:
        if self.ventilation_inlets:
            return self.ventilation_inlets[-1]
        return None

    def get_last_ventilation_inlet(self) -> Optional["System"]:
        if self.ventilation_inlets:
            return self.ventilation_inlets[0]
        return None

    def get_ventilation_outlet(self) -> Optional["System"]:
        if self.ventilation_outlets:
            return self.ventilation_outlets[0]
        return None

    def get_last_ventilation_outlet(self) -> Optional["System"]:
        if self.ventilation_outlets:
            return self.ventilation_outlets[-1]
        return None

    def get_neighhors(self, graph: Graph) -> None:

        neighbors = list(graph.neighbors(self))  # type: ignore
        self.boundaries = []
        windowed_wall_parameters = WindowedWallParameters.from_neighbors(neighbors)
        for wall in [ExternalWall, BaseWindow, InternalElement, FloorOnGround]:
            if wall == ExternalWall:
                self.boundaries.append(
                    VerticalWallParameters.from_neighbors_(
                        neighbors,
                        wall,  # type: ignore
                    )
                )
                self.boundaries.append(
                    RoofWallParameters.from_neighbors_(
                        neighbors,
                        wall,  # type: ignore
                    )
                )
                self.boundaries.append(
                    ExternalWallParameters.from_neighbors_(
                        neighbors,
                        wall,  # type: ignore
                    )
                )

            self.boundaries.append(
                WallParameters.from_neighbors(
                    neighbors,
                    wall,  # type: ignore
                    filter=windowed_wall_parameters.included_external_walls,
                )
            )
        self.boundaries += [windowed_wall_parameters]
        self.boundary_parameters = BoundaryParameters.from_boundaries(self.boundaries)

    def __add__(self, other: "BaseSpace") -> "BaseSpace":
        self.name = (
            f"merge_{self.name.replace('merge', '')}_{other.name.replace('merge', '')}"
        )
        self.volume: float = self.volume + other.volume
        self.external_boundaries += other.external_boundaries
        return self


class Space(BaseSpace):
    def add_to_network(self, network: "Network") -> None:
        network.add_node(self)
        if network.library.merged_external_boundaries:
            external_boundaries = self.merged_external_boundaries
        else:
            external_boundaries = self.external_boundaries  # type: ignore
        for boundary in external_boundaries:
            network.add_node(boundary)
            network.graph.add_edge(
                self,
                boundary,
            )
        emission = self.find_emission()
        if emission:
            network.add_node(emission)
            network.graph.add_edge(
                self,
                emission,
            )
            network._add_subsequent_systems(self.emissions)
        if self.occupancy:
            network.add_node(self.occupancy)
            network.connect_system(self, self.occupancy)
        # Assumption: first element always the one connected to the space.
        if self.get_ventilation_inlet():
            network.add_node(self.get_ventilation_inlet())  # type: ignore
            network.graph.add_edge(self.get_ventilation_inlet(), self)
        if self.get_ventilation_outlet():
            network.add_node(self.get_ventilation_outlet())  # type: ignore
            network.graph.add_edge(self, self.get_ventilation_outlet())
        # The rest is connected to each other
        network._add_subsequent_systems(self.ventilation_outlets)
        network._add_subsequent_systems(self.ventilation_inlets)
        self.assign_position()

    def processing(self, network: "Network") -> None:
        from trano.elements import VAVControl

        _neighbors = []
        if self.get_last_ventilation_inlet():
            _neighbors += list(
                network.graph.predecessors(self.get_last_ventilation_inlet())  # type: ignore
            )
        if self.get_last_ventilation_outlet():
            _neighbors += list(
                network.graph.predecessors(self.get_last_ventilation_outlet())  # type: ignore
            )
        neighbors = list(set(_neighbors))
        controllable_ventilation_elements = list(
            filter(
                None,
                [
                    _get_controllable_element(self.ventilation_inlets),
                    _get_controllable_element(self.ventilation_outlets),
                ],
            )
        )
        for controllable_element in controllable_ventilation_elements:
            if controllable_element.control and isinstance(
                controllable_element.control, VAVControl
            ):
                controllable_element.control.ahu = next(
                    (n for n in neighbors if isinstance(n, AirHandlingUnit)), None
                )

        self.get_neighhors(network.graph)



from enum import Enum

from typing import List, Literal, Optional

from pydantic import BaseModel, Field

DynamicTemplateCategories = Literal["ventilation", "control", "fluid", "boiler"]
SystemContainerTypes = Literal[
    "envelope", "distribution", "emission", "production", "ventilation"
]
ContainerTypes = Literal[SystemContainerTypes, "bus", "solar"]
Pattern = Literal["Solid", "Dot", "Dash", "DashDot"]

TILT_MAPPING = {"wall": 90, "ceiling": 0, "floor": 180}


class Tilt(Enum):
    wall = "wall"
    ceiling = "ceiling"
    floor = "floor"


class Azimuth:
    north = 0
    south = 90
    east = 45
    west = 135


class Flow(str, Enum):
    inlet = "inlet"
    outlet = "outlet"
    radiative = "radiative"
    convective = "convective"
    inlet_or_outlet = "inlet_or_outlet"
    undirected = "undirected"
    interchangeable_port = "interchangeable_port"


class Medium(str, Enum):
    fluid = "fluid"
    heat = "heat"
    data = "data"
    current = "current"
    weather_data = "weather_data"


Boolean = Literal["true", "false"]


class Line(BaseModel):
    template: str
    key: Optional[str] = None
    color: str = "grey"
    label: str
    line_style: str = "solid"
    line_width: float = 1.5


class Axis(BaseModel):
    lines: List[Line] = Field(default=[])
    label: str


class ConnectionView(BaseModel):
    color: Optional[str] = "{255,204,51}"
    thickness: float = 0.1
    disabled: bool = False
    pattern: Pattern = "Solid"


class BaseVariant:
    default: str = "default"



import re
from pathlib import Path
from typing import Any, Dict, List

import yaml

# HOW TO RUN # : ERA001
# material_folders = [ # : ERA001
#     Path("../libraries/IDEAS/IDEAS/Buildings/Data/Materials"), # noqa : ERA001
#     Path("../libraries/IDEAS/IDEAS/Buildings/Data/Insulation"), # noqa : ERA001
#     Path("../libraries/IDEAS/IDEAS/Buildings/Data/Frames"), # noqa : ERA001
# ] # : ERA001
# glazing_folder = Path("../libraries/IDEAS/IDEAS/Buildings/Data/Glazing") # noqa : ERA001
# construction_folder = Path( # : ERA001
#     "../libraries/IDEAS/IDEAS/Buildings/Data/Constructions" # noqa : ERA001
# ) # : ERA001
# parser(material_folders, glazing_folder, construction_folder) # noqa : ERA001


def parse_materials(input_str: str) -> List[Dict[str, Any]]:
    pattern = re.compile(
        r"(IDEAS\.Buildings\.Data\.Materials\.\w+|Materials\.\w+)"
        r"\(d=([\d.]+)(?:,epsLw_a=([\d.]+))?(?:,epsLw_b=([\d.]+))?\)"
    )

    matches = pattern.findall(input_str)
    if not matches:
        raise ValueError("Input string does not match the expected format")

    materials = []
    for match in matches:
        material_type = match[0].split(".")[-1].upper() + ":001"
        thickness = float(match[1])
        type = "glass" if "GLASS" in material_type else "gas"
        material_dict = {
            type: material_type,
            "thickness": thickness,
        }
        materials.append(material_dict)

    return materials


def parse_gas_and_material(input_str: str) -> Dict[str, Any]:
    pattern = r"record\s+(.*?)\s*\""
    match = re.search(pattern, input_str, re.DOTALL)

    if not match:
        raise ValueError("Input string does not match the expected format")
    text = match.group(1).replace("\n", "").replace("\r", "").replace(" ", "")

    pattern_material = re.compile(
        r"(\w+)\s*=\s*IDEAS\.Buildings\.Data\.Interfaces\.(Insulation|Material)\s*\(\s*"
        r"k\s*=\s*([\d.]+)\s*,\s*"
        r"c\s*=\s*([\d.]+)\s*,\s*"
        r"rho\s*=\s*([\d.]+)\s*"
        r"(?:,\s*epsLw\s*=\s*([\d.]+))?\s*"
        r"(?:,\s*epsSw\s*=\s*([\d.]+))?\s*"
        r"(?:,\s*gas\s*=\s*(true|false))?\s*"
        r"(?:,\s*nu\s*=\s*([\d.eE\*]+))?\s*\)"
    )
    pattern_gas = re.compile(
        r"(\w+)\s*=\s*IDEAS\.Buildings\.Data\.Interfaces\.(Insulation|Material)\s*\(\s*"
        r"k\s*=\s*([\d.]+)\s*,\s*"
        r"c\s*=\s*([\d.]+)\s*,\s*"
        r"rho\s*=\s*([\d.]+)\s*"
        r"(?:,\s*epsSw\s*=\s*([\d.]+))?\s*"
        r"(?:,\s*epsLw\s*=\s*([\d.]+))?\s*"
        r"(?:,\s*gas\s*=\s*([\s\S]*))?\s*"
        r"(?:,\s*nu\s*=\s*([\s\S]*))?\s*\)"
    )

    matches = pattern_material.findall(text)
    if not matches:
        matches = pattern_gas.findall(text)
    if not matches:
        raise ValueError("Input string does not match the expected format")

    materials = []
    for match in matches:
        material_type = match[0].upper() + ":001"  # type: ignore
        thermal_conductivity = float(match[2])  # type: ignore
        specific_heat_capacity = float(match[3])  # type: ignore
        density = float(match[4])  # type: ignore
        longwave_emissivity = float(match[5]) if match[5] else None  # type: ignore
        shortwave_emissivity = float(match[6]) if match[6] else None  # type: ignore
        material_dict = {
            "id": material_type,
            "thermal_conductivity": thermal_conductivity,
            "specific_heat_capacity": specific_heat_capacity,
            "density": density,
        }
        if longwave_emissivity is not None:
            material_dict.update({"longwave_emissivity": longwave_emissivity})
        if shortwave_emissivity is not None:
            material_dict.update({"shortwave_emissivity": shortwave_emissivity})
        materials.append(material_dict)
    return materials[0]


def material_parser(folders: List[Path]) -> Dict[str, Any]:
    mo_files = [file for folder in folders for file in list(folder.glob("**/*.mo"))]
    materials: Dict[str, List[Dict[str, Any]]] = {
        "gas": [],
        "material": [],
        "glass_material": [],
    }
    for file in mo_files:
        try:
            results = parse_gas_and_material(file.read_text())
            if results["id"] == "GLASS:001":
                materials["glass_material"].append(results)
            elif any(g in results["id"] for g in ["AIR", "ARGON", "XENON", "KRYPTON"]):
                materials["gas"].append(results)
            else:
                materials["material"].append(results)
        except Exception as e:  # noqa: PERF203
            print(f"Material {file.stem} cannot be generated. Reason {e}")
    return materials


def parse_glazing(input_str: str) -> Dict[str, Any]:
    pattern = re.compile(
        r"record\s+(\w+)\s*=\s*IDEAS\.Buildings\.Data\.Interfaces\.Glazing\s*\(\s*"
        r"final\s+nLay\s*=\s*(\d+),\s*"
        r"(?:final\s+checkLowPerformanceGlazing\s*=\s*(true|false),\s*)?"
        r"final\s+mats\s*=\s*\{([\s\S]*?)\},\s*"
        r"final\s+SwTrans\s*=\s*\[([\s\S]*?)\],\s*"
        r"final\s+SwAbs\s*=\s*\[([\s\S]*?)\],\s*"
        r"final\s+SwTransDif\s*=\s*([\d.]+),\s*"
        r"final\s+SwAbsDif\s*=\s*\{([\s\S]*?)\},\s*"
        r"final\s+U_value\s*=\s*([\d.]+),\s*"
        r"final\s+g_value\s*=\s*([\d.]+)\s*\)"
    )

    match = pattern.search(input_str)
    if not match:
        raise ValueError("Input string does not match the expected format")

    record_id = match.group(1)
    n_lay = int(match.group(2))
    check_low_performance_glazing = match.group(3) == "true" if match.group(3) else None
    mats = match.group(4).replace("\n", "").replace(" ", "")
    sw_trans = match.group(5).replace("\n", "").replace(" ", "")
    sw_abs = match.group(6).replace("\n", "").replace(" ", "")
    sw_trans_dif = float(match.group(7))
    sw_abs_dif = match.group(8).replace("\n", "").replace(" ", "")
    u_value = float(match.group(9))
    g_value = float(match.group(10))

    glazing_dict = {
        "id": record_id,
        "nLay": n_lay,
        "checkLowPerformanceGlazing": check_low_performance_glazing,
        "mats": parse_materials(mats),
        "SwTrans": sw_trans,
        "SwAbs": sw_abs,
        "SwTransDif": sw_trans_dif,
        "SwAbsDif": sw_abs_dif,
        "U_value": u_value,
        "g_value": g_value,
    }

    return glazing_dict


def glazing_parser(folder: Path) -> List[Dict[str, Any]]:
    mo_files = list(folder.glob("*.mo"))
    glazing = []
    for f in mo_files:
        try:
            record = parse_glazing(f.read_text())
            glazing.append(
                {"layers": record["mats"], "id": record["id"].upper() + ":001"}
            )
        except Exception as e:  # noqa: PERF203
            print(f"Glazing {f.stem} cannot be generated. Reason {e}")
    return glazing


def parse_constructions(input_str: str) -> List[Dict[str, Any]]:
    pattern = re.compile(
        r'record\s+(\w+)\s*"([^"]*)"\s*'
        r"extends\s+IDEAS\.Buildings\.Data\.Interfaces\.Construction\s*\(\s*"
        r"(?:locGain\s*=\s*\{[\d,]*\}\s*,\s*)?"
        r"(?:incLastLay\s*=\s*[\w\.]+\s*,\s*)?"
        r"mats\s*=\s*\{([^}]*)\}\s*\)\s*;"
    )

    material_pattern = re.compile(
        r"IDEAS\.Buildings\.Data\.(Materials|Insulation)\.(\w+)\(d=([\d.]+)\)"
    )

    matches = pattern.findall(input_str)
    if not matches:
        raise ValueError("Input string does not match the expected format")

    constructions = []
    for match in matches:
        record_name = match[0]
        materials_str = match[2]

        materials = []
        material_matches = material_pattern.findall(materials_str)
        for material_match in material_matches:
            material_name = material_match[1]
            thickness = float(material_match[2])
            materials.append(
                {"material": material_name.upper() + ":001", "thickness": thickness}
            )

        construction_dict = {
            "id": record_name.upper() + ":001",
            "layers": materials,
        }
        constructions.append(construction_dict)

    return constructions


def construction_parser(folder: Path) -> List[Dict[str, Any]]:
    mo_files = list(folder.glob("*.mo"))
    constructions = []
    for f in mo_files:
        try:
            record = parse_constructions(f.read_text())
            constructions.extend(record)
        except Exception as e:  # noqa: PERF203
            print(f"Construction cannot be generated. Reason {e}")
    return constructions


def parser(
    material_folders: List[Path], glazing_folder: Path, construction_folder: Path
) -> None:
    materials = material_parser(material_folders)
    constructions = construction_parser(construction_folder)
    glazing = glazing_parser(glazing_folder)
    default = {"constructions": constructions, "glazings": glazing, **materials}
    p = Path(__file__).parents[1].joinpath("data/default.yaml")
    with p.open(mode="w+") as f:
        yaml.safe_dump(default, f)



from typing import TYPE_CHECKING

import jinja2.exceptions
from jinja2 import Environment
from pydantic import BaseModel, Field

from trano.elements.types import Axis

if TYPE_CHECKING:
    from trano.elements.base import BaseElement


class Figure(BaseModel):
    right_axis: Axis = Field(default=Axis(lines=[], label=""))
    left_axis: Axis = Field(default=Axis(lines=[], label=""))

    def render_key(self, element: "BaseElement") -> "Figure":
        environment = Environment(autoescape=True)
        for axis in self.right_axis.lines + self.left_axis.lines:
            template = environment.from_string(axis.template)
            try:
                axis.key = template.render(element=element)
            except jinja2.exceptions.UndefinedError:
                continue

        return self


class NamedFigure(Figure):
    name: str



import itertools
from pathlib import Path
from typing import List, Optional, TYPE_CHECKING, Dict

import pandas as pd
from pydantic import BaseModel, Field, computed_field
from networkx.classes.reportviews import NodeView

from trano.elements.base import BaseElement, Control
from trano.elements.data_bus.inputs import BaseInputOutput
from trano.elements.types import ContainerTypes

if TYPE_CHECKING:
    from trano.topology import Network


class ValidationData(BaseModel):
    data: Optional[str] = None
    columns: List[str] = Field([])


class DataBus(BaseElement):
    name: str = "data_bus"
    spaces: List[str] = Field(default=[])
    non_connected_ports: List[BaseInputOutput] = Field(default=[])
    power_ports: List[BaseInputOutput] = Field(default=[])
    external_data: Optional[Path] = None
    container_type: ContainerTypes = "bus"

    @computed_field
    def validation_data(self) -> ValidationData:
        if not self.external_data:
            return ValidationData()
        return transform_csv_to_table(self.external_data)

    def add_to_network(self, network: "Network") -> None:
        from trano.elements import Space, AirHandlingUnit

        spaces = sorted(
            [node for node in network.graph.nodes if isinstance(node, Space)],
            key=lambda x: x.name,
        )
        controls = sorted(
            [node for node in network.graph.nodes if isinstance(node, Control)],
            key=lambda x: x.name,  # type: ignore
        )
        ahus = sorted(
            [node for node in network.graph.nodes if isinstance(node, AirHandlingUnit)],
            key=lambda x: x.name,  # type: ignore
        )
        if not spaces:
            raise ValueError("No spaces in the network")
        self.spaces = [space.name for space in spaces]
        self.external_data = self.external_data
        self.position.set_container(0, 0)
        network.add_node(self)
        for space in spaces:
            network.graph.add_edge(space, self)
            if space.occupancy:
                network.graph.add_edge(space.occupancy, self)
        for control in controls:
            network.graph.add_edge(control, self)
        for ahu in ahus:
            network.graph.add_edge(ahu, self)

    def configure(self, network: "Network") -> None:
        self.non_connected_ports = get_non_connected_ports(network.graph.nodes)
        self.power_ports = get_power_ports(network.graph.nodes)


def get_power_ports(nodes: List[NodeView]) -> List[BaseInputOutput]:
    power_ports = []
    for node in nodes:
        if not (
            hasattr(node, "component_template")
            and hasattr(node.component_template, "bus")
        ):
            continue
        if node.component_template and node.component_template.bus:
            node_ports = node.component_template.bus.list_ports(node)
            power_ports += [p for p in node_ports["RealOutput"] if p.power is not None]
    return power_ports


def get_non_connected_ports(nodes: List[NodeView]) -> List[BaseInputOutput]:
    port_types = ["Real", "Integer", "Boolean"]
    ports: Dict[str, List[BaseInputOutput]] = {
        f"{port_type}{direction}": []
        for port_type in port_types
        for direction in ["Output", "Input"]
    }

    for node in nodes:
        if not (
            hasattr(node, "component_template")
            and hasattr(node.component_template, "bus")
        ):
            continue
        if node.component_template and node.component_template.bus:
            node_ports = node.component_template.bus.list_ports(node)
            for port_type in port_types:
                ports[f"{port_type}Output"] += node_ports[f"{port_type}Output"]
                ports[f"{port_type}Input"] += node_ports[f"{port_type}Input"]

    for port_type in port_types:
        ports[f"{port_type}Output"] = list(set(ports[f"{port_type}Output"]))
        ports[f"{port_type}Input"] = list(set(ports[f"{port_type}Input"]))

    return list(
        itertools.chain(
            *[
                _get_non_connected_ports_intersection(
                    ports[f"{port_type}Input"], ports[f"{port_type}Output"]
                )
                for port_type in port_types
            ]
        )
    )


def _get_non_connected_ports_intersection(
    input_ports: List[BaseInputOutput], output_ports: List[BaseInputOutput]
) -> List[BaseInputOutput]:
    return list(set(input_ports) - set(output_ports).intersection(set(input_ports)))


def transform_csv_to_table(
    file_path: Path, total_second: bool = True
) -> ValidationData:
    data = pd.read_csv(file_path, index_col=0, infer_datetime_format=True)
    data = data.ffill().bfill()
    data = data.dropna(axis=1)
    data.index = pd.to_datetime(data.index)
    if total_second:
        data.index = (data.index - data.first_valid_index()).total_seconds()  # type: ignore
    else:
        data.index = data.index.astype(int) // 10**9
    data_str = data.to_csv(sep=",", header=False, lineterminator=";")
    if data_str.endswith(";"):
        data_str = data_str[:-1]
    return ValidationData(data=data_str, columns=data.columns.tolist())



from pathlib import Path
from typing import Any, Dict, Optional, Type

import yaml
from pydantic import ConfigDict, Field, computed_field, create_model

from trano.elements.common_base import BaseParameter
from trano.elements.utils import _get_default, _get_type


def load_parameters() -> Dict[str, Type["BaseParameter"]]:
    # TODO: remove absoluth path reference
    parameter_path = (
        Path(__file__).parents[2].joinpath("data_models", "parameters.yaml")
    )
    data = yaml.safe_load(parameter_path.read_text())
    classes = {}

    for name, parameter in data.items():
        attrib_ = {}
        for k, v in parameter["attributes"].items():
            alias = v.get("alias", None)
            alias = alias if alias != "None" else None
            if v.get("range"):
                attrib_[k] = (
                    _get_type(v["range"]),
                    Field(
                        default=_get_default(v),
                        alias=alias,
                        description=v.get("description", None),
                    ),
                )
            else:
                attrib_[k] = computed_field(  # type: ignore # TODO: why?
                    eval(v["func"]),  # noqa: S307
                    return_type=eval(v["type"]),  # noqa: S307
                    alias=alias,  # TODO: avoid using eval
                )
        model = create_model(f"{name}_", __base__=BaseParameter, **attrib_)  # type: ignore # TODO: why?
        if parameter["classes"] is None:
            continue
        for class_ in parameter["classes"]:
            classes[class_] = model
    return classes


PARAMETERS = load_parameters()


def param_from_config(name: str) -> Optional[Type[BaseParameter]]:
    if name in PARAMETERS:
        return PARAMETERS[name]
    elif name.upper() in PARAMETERS:
        return PARAMETERS[name.upper()]
    else:
        return None
    # TODO: to be replaced with a raise later


def change_alias(
    parameter: BaseParameter, mapping: Optional[Dict[str, str]] = None
) -> Any:  # noqa: ANN401
    mapping = mapping or {}
    new_param = {}
    for name, field in parameter.model_fields.items():
        if mapping.get(name):
            field.alias = mapping[name]
        new_param[name] = (
            field.annotation,
            Field(field.default, alias=field.alias, description=field.description),
        )

    for name, field in parameter.model_computed_fields.items():  # type: ignore
        if mapping.get(name):
            new_param[name] = (
                Optional[field.return_type],  # type: ignore
                Field(None, alias=mapping[name], description=field.description),
            )
    return create_model(  # type: ignore
        "new_model",
        **new_param,
        __config__=ConfigDict(populate_by_name=True),
    )


def modify_alias(
    parameter: BaseParameter, modify_alias: Dict[str, str]
) -> Any:  # noqa: ANN401
    return change_alias(parameter, modify_alias)(**parameter.model_dump()).model_dump(
        by_alias=True, include=set(modify_alias)
    )


def exclude_parameters(
    parameters: BaseParameter, exclude_parameters: Optional[set[str]] = None
) -> Dict[str, Any]:
    return parameters.model_dump(by_alias=True, exclude=exclude_parameters)


def default_parameters(parameters: BaseParameter) -> Dict[str, Any]:
    if not parameters:
        return {}
    return parameters.model_dump(by_alias=True, exclude_none=True)



from pathlib import Path
from typing import Any, Dict, List, Union

import docx
from buildingspy.io.outputfile import Reader  # type: ignore
from docx import Document
from docx.enum.text import WD_ALIGN_PARAGRAPH
from docx.oxml import parse_xml
from docx.oxml.ns import nsdecls
from docx.shared import Pt
from docx.table import _Cell

from trano.plot.plot import add_element_figures
from trano.reporting.reporting import ModelDocumentation
from trano.reporting.types import BaseNestedTable, Topic
from trano.reporting.utils import get_description

COLUMN_SIZE_WITH_DESCRIPTION = 3

TABLE_COUNT = 1


def _round(value: float | str | None) -> str:
    try:
        return str(round(value, 4))  # type: ignore
    except TypeError:
        return str(value)


def _set_color(cell: _Cell, key: str) -> None:
    if key == "type":
        cell._tc.get_or_add_tcPr().append(
            parse_xml(rf'<w:shd {nsdecls("w")} w:fill="#D3D3D3"/>')
        )
    if key == "name":
        cell._tc.get_or_add_tcPr().append(
            parse_xml(rf'<w:shd {nsdecls("w")} w:fill="#FF7F7F"/>')
        )
    if key == "thickness":
        cell._tc.get_or_add_tcPr().append(
            parse_xml(rf'<w:shd {nsdecls("w")} w:fill="#FFFFE0"/>')
        )


def create_table(doc: Document, data: Dict[str, Any]) -> None:  # type: ignore
    size = 2
    description = get_description()
    if any(key in description for key in data):
        size = COLUMN_SIZE_WITH_DESCRIPTION
    table = doc.add_table(cols=size, rows=len(data))  # type: ignore
    for i, (key, value) in enumerate(data.items()):
        table.rows[i].cells[0].text = key
        _set_color(table.rows[i].cells[0], key)
        if isinstance(value, dict):
            create_table(table.rows[i].cells[1], value)
        elif isinstance(value, list) and all(isinstance(v, dict) for v in value):
            for v in value:
                create_table(table.rows[i].cells[1], v)
        else:
            _set_color(table.rows[i].cells[1], key)
            table.rows[i].cells[1].text = _round(value)
            if size == COLUMN_SIZE_WITH_DESCRIPTION:
                table.rows[i].cells[2].text = str(description.get(key, "N/A"))


def add_table_caption(doc: Document, caption_text: str) -> None:  # type: ignore
    global TABLE_COUNT  # noqa: PLW0603
    caption = doc.add_paragraph(f"Table {TABLE_COUNT}: {caption_text}", style="Caption")  # type: ignore
    TABLE_COUNT += 1
    caption.alignment = WD_ALIGN_PARAGRAPH.LEFT
    caption.space_after = Pt(0)


def create_tables_and_figures(
    doc: Document,  # type: ignore
    data: Union[List[BaseNestedTable], BaseNestedTable],
    topic: Topic,
    documentation: ModelDocumentation,
) -> None:
    if isinstance(data, list):
        for d in data:
            if d.name:
                add_table_caption(doc, f"Characteristics of {topic} {d.name}.")
                create_table(doc, d.model_dump())
                doc.add_paragraph()  # type: ignore
                insert_figure(doc, d.name, documentation)
                doc.add_paragraph()  # type: ignore

    elif data.name:
        add_table_caption(doc, f"Characteristics of {topic} {data.name}.")
        create_table(doc, data.model_dump())
        doc.add_paragraph()  # type: ignore
        insert_figure(doc, data.name, documentation)
        doc.add_paragraph()  # type: ignore


def insert_figure(doc: Document, element_name: str, documentation: ModelDocumentation) -> None:  # type: ignore
    if documentation.result is None:
        return
    mat = Reader(documentation.result.path, documentation.result.type)
    elements = [node for node in documentation.elements if node.name == element_name]
    if elements:
        element = elements[0]
        add_element_figures(doc, mat, element)


def to_docx(documentation: ModelDocumentation, path: Path) -> docx.document.Document:
    document = Document()
    document.add_page_break()  # type: ignore
    for doc in [
        documentation.spaces,
        documentation.systems,
        documentation.constructions,
    ]:
        document.add_heading(doc.topic, level=2)  # type: ignore
        if doc.introduction:
            document.add_paragraph(doc.introduction)
        create_tables_and_figures(document, doc.table, doc.topic, documentation)  # type: ignore
        if doc.conclusions:
            document.add_paragraph(doc.conclusions)
        document.add_page_break()  # type: ignore
    document.save(str(path))
    return document



from functools import cache
from pathlib import Path
from typing import Any, ClassVar, Dict, List, Optional, Type, TYPE_CHECKING, get_args

from jinja2 import Environment, FileSystemLoader
from pydantic import (
    ConfigDict,
    Field,
    field_validator,
    model_validator,
    PrivateAttr,
)

from trano.elements.common_base import (
    BaseElementPosition,
    ComponentModel,
    MediumTemplate,
)
from trano.elements.common_base import BaseParameter
from trano.elements.connection import Port
from trano.elements.figure import NamedFigure
from trano.elements.library.base import DynamicComponentTemplate, LibraryData
from trano.elements.library.parameters import param_from_config
from trano.elements.types import BaseVariant, ContainerTypes
from trano.exceptions import UnknownComponentVariantError

if TYPE_CHECKING:
    from trano.topology import Network
    from trano.elements.library.library import Library


class BaseElementPort(BaseElementPosition):
    name: Optional[str] = Field(default=None)
    ports: list[Port] = Field(default=[], validate_default=True)
    container_type: Optional[ContainerTypes] = None

    def available_ports(self) -> List[Port]:
        return [port for port in self.ports if not port.connected]

    def reset_port_counters(self) -> None:
        for port in self.ports:
            port.reset_counter()


def default_environment() -> Environment:
    environment = Environment(
        trim_blocks=True,
        lstrip_blocks=True,
        loader=FileSystemLoader(str(Path(__file__).parents[1].joinpath("templates"))),
        autoescape=True,
    )
    environment.filters["enumerate"] = enumerate
    return environment


class BaseElement(BaseElementPort):
    name_counter: ClassVar[int] = (
        0  # TODO: this needs to be removed and replaced with a proper solution.
    )

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="allow")
    medium: MediumTemplate = Field(default_factory=MediumTemplate)
    parameters: Optional[BaseParameter] = None
    template: Optional[str] = None
    component_template: Optional[DynamicComponentTemplate] = None
    variant: str = BaseVariant.default
    libraries_data: List[LibraryData] = Field(default=[])
    figures: List[NamedFigure] = Field(default=[])
    _environment: Environment = PrivateAttr(default_factory=default_environment)
    component_model: Optional[ComponentModel] = None

    @model_validator(mode="before")
    @classmethod
    def validate_libraries_data(cls, value: Dict[str, Any]) -> Dict[str, Any]:
        if isinstance(value, dict):
            if "libraries_data" not in value:
                from trano.elements.library.components import COMPONENTS

                value["libraries_data"] = COMPONENTS.get_components(cls.__name__)
            parameter_class = param_from_config(cls.__name__)
            if (
                parameter_class
                and isinstance(value, dict)
                and not value.get("parameters")
            ):
                value["parameters"] = parameter_class()

        return value

    @model_validator(mode="after")
    def assign_default_name(self) -> "BaseElement":
        if self.name is None:
            self.name = f"{type(self).__name__.lower()}_{type(self).name_counter}"
            type(self).name_counter += 1
        return self

    @field_validator("name")
    @classmethod
    def clean_name(cls, value: str) -> str:
        if ":" in value:
            return value.lower().replace(":", "_")
        return value

    def get_library_data(self, library: "Library") -> Optional[LibraryData]:
        libraries_data_variants = [
            library_
            for library_ in self.libraries_data
            if library_.variant == self.variant
        ]
        if not libraries_data_variants:
            raise UnknownComponentVariantError(
                f"Library data not found for {self.name} in {library.name} for variant {self.variant}"
            )
        libraries_data = [
            library_
            for library_ in libraries_data_variants
            if (library_.library == library.name.lower())
        ]
        if not libraries_data:
            libraries_data = [
                library_
                for library_ in libraries_data_variants
                if (library_.library == "default")
            ]
        if libraries_data:
            return libraries_data[0]
        return None

    def assign_library_property(self, library: "Library") -> bool:
        if library.medium.is_empty():
            raise ValueError("Library medium is empty")
        if not self.libraries_data:
            return False
        library_data = self.get_library_data(library)
        if self.medium.is_empty():
            self.medium = library.medium
        if not library_data:
            return False
        if not self.ports:
            self.ports = library_data.ports()
        if not self.template:
            self.template = library_data.template
        if not self.component_template:
            self.component_template = library_data.component_template
        if not self.figures and library_data.figures:
            self.figures = [
                NamedFigure(**(fig.render_key(self).model_dump() | {"name": self.name}))
                for fig in library_data.figures
            ]

        return True

    def processed_parameters(self, library: "Library") -> Any:  # noqa: ANN401
        if self.libraries_data:
            library_data = self.get_library_data(library)
            if library_data and self.parameters:
                return library_data.parameter_processing(self.parameters)
        return {}

    def set_position(self, layout: Dict[str, Any], global_: bool = True) -> None:
        if (
            self.position.is_global_empty()
            if global_
            else self.position.is_container_empty()
        ):
            x, y = list(layout.get(self.name))  # type: ignore
            (
                self.position.set_global(x, y)
                if global_
                else self.position.set_container(x, y)
            )

    def get_controllable_ports(self) -> List[Port]:
        return [port for port in self.ports if port.is_controllable()]

    @property
    def type(self) -> str:
        return type(self).__name__

    @cache  # noqa: B019
    def model(self, network: "Network") -> Optional[ComponentModel]:
        if not self.template:
            return None
        self._environment.globals.update(network.library.functions)
        if self.component_template:
            component = self.component_template.render(
                network.name, self, self.processed_parameters(network.library)
            )
            if self.component_template.category:
                network.dynamic_components[self.component_template.category].append(
                    component
                )
        package_name = network.name
        library_name = network.library.base_library()
        parameters = self.processed_parameters(network.library)
        component_model: Dict[str, Any] = {"id": hash(self)}
        for model_type, annotation in {
            "model": self.position.global_.annotation,
            "container": self.position.container.annotation,
        }.items():
            template = self._environment.from_string(
                "{% import 'macros.jinja2' as macros %}"
                + self.template
                + " "
                + annotation
            )
            component_model.update(
                {
                    model_type: template.render(
                        element=self,
                        package_name=package_name,
                        library_name=library_name,
                        parameters=parameters,
                    )
                }
            )
        component_model_ = ComponentModel.model_validate(component_model)
        self.component_model = component_model_
        return component_model_

    def __hash__(self) -> int:
        return hash(f"{self.name}-{type(self).__name__}")

    def add_to_network(self, network: "Network") -> None:
        network.add_node(self)

    def processing(self, network: "Network") -> None: ...

    def configure(self, network: "Network") -> None: ...
    def system_ports_connected(self) -> bool:
        return True

    def non_connected_ports_name(self) -> List[str]:
        return [
            f"{self.name}.{n}"
            for port in self.ports
            if not port.connected
            for n in port.names
        ]

    def assign_container_type(self, network: "Network") -> None:
        if self.container_type is None:
            network.graph.neighbors(self)
            container_types = {
                s.container_type
                for s in list(network.graph.predecessors(self))  # type: ignore
                + list(network.graph.successors(self))  # type: ignore
                if s.container_type
            }
            self.container_type = next(
                iter(
                    sorted(
                        container_types,
                        key=lambda type_: {v: i for i, v in enumerate(get_args(ContainerTypes))}.get(type_),  # type: ignore
                    )
                )
            )


class ElementPort(BaseElementPort):
    element_type: Optional[Type[BaseElement]] = None
    merged_number: int = 1

    def has_target(self, targets: List[Type[BaseElement]]) -> bool:
        return (not targets) or bool(
            self.element_type is not None
            and targets
            and any(issubclass(self.element_type, t) for t in targets)
        )

    def get_connection_per_target(
        self, target: Optional[Type[BaseElement]] = None
    ) -> int:
        if target is None:
            return 0
        return len(
            [
                target_
                for port in self.ports
                for target_ in port.targets
                if issubclass(target, target_)
            ]
        )

    @classmethod
    def from_element_without_ports(cls, element: BaseElement) -> "ElementPort":
        return cls.from_element(element, use_original_ports=False)

    @classmethod
    def from_element(
        cls, element: BaseElement, use_original_ports: bool = True
    ) -> "ElementPort":
        from trano.elements.envelope import MergedBaseWall

        merged_number = 1
        if isinstance(element, MergedBaseWall):
            merged_number = len(element.surfaces)
        if element.position is not None:
            element_port = cls(
                **(
                    element.model_dump()
                    | {"element_type": type(element), "merged_number": merged_number}
                )
            )
        else:
            element_port = cls(
                **(
                    element.model_dump(exclude={"position"})
                    | {"element_type": type(element), "merged_number": merged_number}
                )
            )
        if use_original_ports:
            element_port.ports = element.ports
        return element_port


# This has to be here for now!!!
class Control(BaseElement):
    controllable_element: Optional[BaseElement] = None
    space_name: Optional[str] = None
    container_annotation_template: str = """annotation (
    Placement(transformation(origin = {{ macros.join_list(element.container_position) }},
    extent = {% raw %}{{5, -5}, {-5, 5}}
    {% endraw %})));"""



import logging
from pathlib import Path
from types import SimpleNamespace
from typing import List, Optional, get_args, Tuple, cast, Set

import yaml
from jinja2 import Environment, FileSystemLoader, Template
from pydantic import BaseModel, Field, model_validator, computed_field, field_validator

from trano.elements import Port, Connection, BaseElement
from trano.elements.base import ElementPort
from trano.elements.common_base import (
    BasePosition,
    Point,
    BaseProperties,
    MediumTemplate,
)
from trano.elements.connection import (
    BasePartialConnectionWithContainerType,
    PartialConnection,
    connect,
    ContainerConnection,
)
from trano.elements.types import (
    ContainerTypes,
    ConnectionView,
    Medium,
    SystemContainerTypes,
)
from trano.elements.utils import wrap_with_raw
from trano.exceptions import ContainerNotFoundError

logger = logging.getLogger(__name__)
ENVIRONMENT = Environment(
    trim_blocks=True,
    lstrip_blocks=True,
    loader=FileSystemLoader(str(Path(__file__).parents[1].joinpath("templates"))),
    autoescape=True,
)
ENVIRONMENT.filters["frozenset"] = frozenset
ENVIRONMENT.filters["enumerate"] = enumerate


class PortGroup(BaseModel):
    connected_container_names: List[ContainerTypes]
    ports: List[Port]


class PortGroupMedium(BaseModel):
    medium: Medium
    port: Port


class BaseContainer(BaseModel):
    component_size: Point = Field(default=Point(x=20, y=20))


class ContainerInput(BaseModel):
    nodes: List[BaseElement]
    connections: List[Connection]
    data: BaseProperties
    medium: MediumTemplate


class ContainerLayout(BaseModel):
    bottom_left: Point = Point(x=-100, y=-100)
    top_right: Point = Point(x=100, y=100)
    global_origin: Point

    @computed_field  # type: ignore[prop-decorator]
    @property
    def scale(self) -> int:
        scale_x = int(self.top_right.c_.x - self.bottom_left.c_.x)
        scale_y = int(self.top_right.c_.y - self.bottom_left.c_.y)
        if scale_x != scale_y:
            raise ValueError("Scale x and y must be equal")
        return scale_x


class Container(BaseContainer):
    name: ContainerTypes
    port_groups: List[PortGroup]
    port_groups_per_medium: List[PortGroupMedium] = Field(default_factory=list)
    connections: List[ContainerConnection] = Field(default_factory=list)
    elements: List[BaseElement] = Field(default_factory=list)
    template: str
    left_boundary: Optional[ContainerTypes] = None
    data: Optional[BaseProperties] = None
    layout: ContainerLayout = Field(
        default_factory=lambda: ContainerLayout(global_origin=Point(x=0, y=0))
    )
    prescribed_connection_equation: str = ""

    @field_validator("template")
    @classmethod
    def _template_validator(cls, value: str) -> str:
        return wrap_with_raw(value)

    def has_data(self) -> bool:
        return self.data is not None

    def set_data(self, data: BaseProperties) -> None:
        if self.name == data.container_type:
            self.data = data

    def contain_elements(self) -> bool:
        return bool(self.elements)

    def get_equation_view(self) -> Set[Tuple[str, ...]]:
        return {c.equation_view() for c in self.connections}

    def main_equation(self) -> str:
        location = f"{self.layout.global_origin.c_.x},{self.layout.global_origin.c_.y}"
        size = (
            f"{self.layout.global_origin.c_.x+self.component_size.c_.x},"
            f"{self.layout.global_origin.c_.y+self.component_size.c_.y}"
        )
        return (
            f"Components.Containers.{self.name} {self.name}1 "
            f"annotation (Placement(transformation(extent={{{{{location}}},"
            f"{{{size}}}}})));"
        )

    @model_validator(mode="after")
    def _validate(self) -> "Container":
        if self.name in [
            container_name
            for port_group in self.port_groups
            for container_name in port_group.connected_container_names
        ]:
            raise ValueError(f"Container {self.name} cannot be connected to itself.")
        return self

    def get_port_group(
        self, connected_container_name: Optional[ContainerTypes] = None
    ) -> Optional[PortGroup]:
        for port_group in self.port_groups:
            if connected_container_name in port_group.connected_container_names:
                return port_group
        return None

    def _initialize_template(self, medium: MediumTemplate) -> None:
        ports = {}
        for port_group in self.port_groups:
            for port in port_group.ports:
                for name in port.names:
                    ports[name] = str(port.counter - 1)
        try:
            template = ENVIRONMENT.from_string(self.template)
        except:
            raise
        self.template = template.render(medium=medium, ports=SimpleNamespace(**ports))

    def build(self, template: Template, medium: MediumTemplate) -> str:
        self._initialize_template(medium)
        self.add_grouped_by_medium_connection()
        return template.render(container=self)

    def add_grouped_by_medium_connection(self) -> None:
        for group_per_medium in self.port_groups_per_medium:
            for element in self.elements:
                for element_port in element.ports:
                    if element_port.medium == group_per_medium.medium:
                        position = BasePosition()
                        position.set(
                            element.position.x_container, element.position.y_container
                        )
                        e1 = ElementPort(
                            ports=[element_port.without_targets()],
                            container_type=self.name,
                            position=position,
                            name=element.name,
                        )
                        e2 = ElementPort(
                            ports=[group_per_medium.port],
                            container_type=self.name,
                            position=position,
                        )
                        connections = connect(e1, e2)
                        if connections:
                            self.connections += [
                                ContainerConnection.model_validate(
                                    c.model_dump()
                                    | {
                                        "source": tuple(
                                            sorted([c.right.equation, c.left.equation])
                                        )
                                    }
                                )
                                for c in connections
                            ]


class MainContainerConnection(BaseModel):
    left: PartialConnection
    right: PartialConnection
    annotation: Optional[str] = None

    @classmethod
    def from_list(
        cls, connections: List[BasePartialConnectionWithContainerType]
    ) -> "MainContainerConnection":
        return cls(left=connections[0], right=connections[1])

    def get_equation(self) -> str:
        return (
            f"connect({self.left.container_type}1.{self.left.equation}, "
            f"{self.right.container_type}1.{self.right.equation})"
        )


class Location(BaseModel):
    point_1: Point
    point_2: Point


class ConnectionList(BaseModel):
    connection_type: List[str]
    annotation: str


class BusConnection(BaseModel):
    connection_type: Tuple[str, str]
    location: str

    @model_validator(mode="after")
    def _validator(self) -> "BusConnection":
        self.connection_type = cast(
            Tuple[str, str], tuple(sorted(self.connection_type))
        )

        return self

    def equation(self) -> str:
        return (
            f"connect({self.connection_type[0]}1.dataBus, {self.connection_type[1]}1.dataBus) "
            f"annotation (Line(points={self.location}, color={{255,204,51}}, thickness=0.5));"
        )


class Containers(BaseModel):
    containers: List[Container]
    main: Optional[str] = None
    connections: List[MainContainerConnection] = Field(default=[])
    bus_connections: List[BusConnection] = [
        BusConnection(
            connection_type=("bus", "envelope"),
            location="{{-83.8,48.8},{-83.8,56},{-60,56},{-60,26},{-74,26},{-74,20}}",
        ),
        BusConnection(
            connection_type=("emission", "envelope"),
            location="{{-95.8,16.8},{-95.8,26},{-55.8,26},{-55.8,16.8}}",
        ),
        BusConnection(
            connection_type=("distribution", "envelope"),
            location="{{-95.8,16.8},{-95.8,26},{-15.8,26},{-15.8,16.8}}",
        ),
        BusConnection(
            connection_type=("production", "envelope"),
            location="{{-95.8,16.8},{-95.8,26},{24.2,26},{24.2,16.8}}",
        ),
        BusConnection(
            connection_type=("ventilation", "envelope"),
            location="{{-44.1,-32.6},{-50,-32.6},{-50,-16},{-90,-16},{-90,15.8},{-83.9,15.8}}",
        ),
    ]
    connection_list: List[ConnectionList] = [
        ConnectionList(
            connection_type=["envelope1.heatPortCon", "emission1.heatPortCon"],
            annotation="""annotation (Line(points={{-64,15},{-62,15.2},{-43.8,15.2}}, color={191,0,0}));""",
        ),
        ConnectionList(
            connection_type=["emission1.heatPortRad", "envelope1.heatPortRad"],
            annotation="""annotation (Line(points
        ={{-64,4.8},{-48,4.8},{-48,5},{-44,5}}, color={191,0,0}));""",
        ),
        ConnectionList(
            connection_type=["distribution1.port_a", "production1.port_b1"],
            annotation="""annotation (Line(points={{15.8,15},{30,15},{30,15},{36,15}},
                                          color={0,127,255}));""",
        ),
        ConnectionList(
            connection_type=["production1.port_a1", "distribution1.port_b"],
            annotation="""annotation (Line(points={{16,5},{32,5},{32,5.2},{36,5.2}},
                                          color={0,127,255}));""",
        ),
        ConnectionList(
            connection_type=["distribution1.port_a1", "emission1.port_b"],
            annotation="""annotation (Line(points={{-24,
          4.6},{-8,4.6},{-8,5.2},{-4,5.2}}, color={0,127,255}));""",
        ),
        ConnectionList(
            connection_type=["emission1.port_a", "distribution1.port_b1"],
            annotation="""annotation (Line(points={{-24,15.4},{-22,15},{-4,15}}, color={0,127,255}));""",
        ),
        ConnectionList(
            connection_type=["envelope1.ports_b", "bus1.port_b"],
            annotation="""annotation (Line(points={{-24,
      5},{-20,5},{-20,10},{-8,10},{-8,14.8},{-4,14.8}}, color={0,127,255}));""",
        ),
        ConnectionList(
            connection_type=["envelope1.y", "bus1.u"],
            annotation="""annotation (Line(points={{-85.7,10.1},{-94,10.1},
          {-94,40},{-85.8,40}}, color={0,0,127}));""",
        ),
        ConnectionList(
            connection_type=["envelope1.heatPortCon1", "bus1.heatPortCon"],
            annotation="""annotation (Line(points={{-24,
    5},{-20,5},{-20,10},{-8,10},{-8,14.8},{-4,14.8}}, color={0,127,255}));""",
        ),
        ConnectionList(
            connection_type=["envelope1.ports_a", "ventilation1.ports_b"],
            annotation="""annotation (Line(points={{-84.1,
          3.4},{-88,3.4},{-88,-14},{-48,-14},{-48,-20.2},{-43.9,-20.2}}, color={
          0,127,255}));""",
        ),
        ConnectionList(
            connection_type=["ventilation1.ports_a", "envelope1.ports_b"],
            annotation="""annotation (Line(points={{-44.1,
          -32.6},{-50,-32.6},{-50,-16},{-90,-16},{-90,15.8},{-83.9,15.8}},
        color={0,127,255}));""",
        ),
        ConnectionList(
            connection_type=["ventilation1.ports_b", "envelope1.ports_b"],
            annotation="""annotation (Line(points={{-44.1,
          -32.6},{-50,-32.6},{-50,-16},{-90,-16},{-90,15.8},{-83.9,15.8}},
        color={0,127,255}));""",
        ),
    ]

    def add_data(self, data: BaseProperties) -> None:
        for container in self.containers:
            container.set_data(data)

    @classmethod
    def load_from_config(cls) -> "Containers":
        config_yaml = (
            Path(__file__).parents[1].joinpath("elements/config/containers.yaml")
        )
        data = yaml.safe_load(config_yaml.read_text())
        return cls.model_validate(data)

    def _set_connection_annotation(self) -> None:
        for connection in self.connections:
            for connection_list in self.connection_list:
                if all(
                    c in connection.get_equation()
                    for c in connection_list.connection_type
                ):
                    connection.annotation = connection_list.annotation

            if connection.annotation is None:
                raise ValueError(
                    f"Connection {connection.get_equation()} not found in connection list."
                )

    @model_validator(mode="after")
    def _validate(self) -> "Containers":
        container_names = [container.name for container in self.containers]
        if len(container_names) != len(set(container_names)):
            raise ValueError("Containers must have unique names.")
        return self

    def build(self, container_input: ContainerInput) -> List[str]:
        self.assign_nodes(container_input.nodes)
        self.connect(container_input.connections)
        self.build_main_connections()
        self.add_data(container_input.data)
        template = self._template()
        main_template = self._main_template()
        self._set_connection_annotation()
        self.main = main_template.render(container=self)

        return [
            c.build(template, container_input.medium) for c in self.in_use_containers()
        ]

    def build_main_connections(self) -> None:
        connections = [
            conn
            for c in self.containers
            for conn in c.connections
            if not conn.in_the_same_container()
        ]
        couple_connections = {
            c.source: [
                c_.get_container_equation()
                for c_ in connections
                if c_.source == c.source
            ]
            for c in connections
        }
        for equations in couple_connections.values():
            if len(equations) == 2:
                self.connections += [MainContainerConnection.from_list(equations)]  # type: ignore

    def _get_connection_view(
        self, connected_container_name: Optional[ContainerTypes] = None
    ) -> ConnectionView:
        connection_view = ConnectionView()
        if connected_container_name in get_args(SystemContainerTypes):
            connection_view = ConnectionView(
                color="{0, 0, 139}", thickness=0.1, pattern="Dash"
            )
        if connected_container_name in ["bus"]:
            connection_view = ConnectionView(color=None, thickness=0.2, disabled=True)
        return connection_view

    def connect(self, connections: List[Connection]) -> None:
        for connection in connections:
            edge_left = connection.left
            edge_right = connection.right
            if edge_left.container_type == edge_right.container_type:
                container = self.get_container(edge_left.container_type)
                if not container:
                    raise ContainerNotFoundError(
                        f"Container {edge_left.container_type} not found."
                    )
                container.connections += [
                    ContainerConnection.model_validate(
                        connection.model_dump()
                        | {
                            "source": tuple(
                                sorted([edge_right.equation, edge_left.equation])
                            )
                        }
                    )
                ]
                continue

            for edge_1, edge_2 in [(edge_left, edge_right), (edge_right, edge_left)]:
                container = self.get_container(edge_1.container_type)
                if container:
                    port_group = container.get_port_group(edge_2.container_type)
                    if port_group:
                        connection_view = self._get_connection_view(
                            edge_2.container_type
                        )
                        container_position = BasePosition()
                        container_position.set(0, 0)
                        element_1 = ElementPort(
                            name=edge_1.name,
                            ports=[
                                edge_1.port.without_targets()
                                .disconnect()
                                .set_ignore_direction()
                                .substract_counter()
                            ],
                            container_type=edge_1.container_type,
                            position=edge_1.position,
                        )
                        element_2 = ElementPort(
                            ports=port_group.ports,
                            container_type=edge_1.container_type,
                            position=container_position,
                        )
                        connections_ = connect(element_1, element_2)
                        if not connections_:
                            logger.debug(
                                f"Element {element_1.name} from container {element_1.container_type} "
                                f"cannot be connected with container {edge_2.container_type}"
                            )
                        container.connections += [
                            ContainerConnection.model_validate(
                                c.model_copy(
                                    update={"connection_view": connection_view}
                                ).model_dump()
                                | {
                                    "source": tuple(
                                        sorted([edge_1.equation, edge_2.equation])
                                    )
                                }
                            )
                            for c in connections_
                        ]

    def assign_nodes(self, nodes: List[BaseElement]) -> None:
        for container_type in get_args(ContainerTypes):
            container = self.get_container(container_type)
            node_types = [
                node for node in nodes if node.container_type == container_type
            ]

            if container:
                container.elements.extend(node_types)

    def in_use_containers(self) -> List[Container]:
        return [c for c in self.containers if c.contain_elements()]

    def bus_equations(self) -> List[str]:
        containers = {container.name for container in self.in_use_containers()}
        return [
            bus_connection.equation()
            for bus_connection in self.bus_connections
            if set(bus_connection.connection_type).issubset(containers)
        ]

    def get_container(
        self, container_type: Optional[ContainerTypes] = None
    ) -> Optional[Container]:
        for container in self.containers:
            if container.name == container_type:
                return container
        return None

    def _template(self) -> Template:
        template = ENVIRONMENT.get_template("containers.jinja2")
        return template

    def _main_template(self) -> Template:
        template_ = """
model building

{% for container_ in container.in_use_containers() %}
{{ container_.main_equation() | safe }}
{% endfor %}
{% raw %}
Buildings.Electrical.AC.OnePhase.Interfaces.Terminal_p term_p
annotation (Placement(transformation(extent={{-126,-18},{-92,18}}),
iconTransformation(
extent={{-112,-12},{-88,12}})));
equation
connect(term_p, bus1.term_p) annotation (Line(points={{-109,0},{-88,0},
        {-88,-10},{60,-10},{60,64},{-50,64},{-50,40},{-65,40}}, color={
        0,120,120}));
{% endraw %}
{% for connection in container.connections %}
connect({{ connection.left.container_type }}1.{{ connection.left.equation }},
{{ connection.right.container_type }}1.{{ connection.right.equation }})
{{ connection.annotation }}
{% endfor %}

{% for bus_equation in container.bus_equations() %}
{{ bus_equation | safe }}
{% endfor %}
{% for container_ in container.in_use_containers() %}
{{ container_.prescribed_connection_equation | safe }}
{% endfor %}
{% raw %}
annotation (Icon(coordinateSystem(preserveAspectRatio=false), graphics={
          Rectangle(
            extent={{-100,100},{100,-100}},
            fillColor={215,215,215},
            fillPattern=FillPattern.Solid,
            pattern=LinePattern.None),
        Rectangle(
          extent={{-74,18},{22,-40}},
            fillColor={255,255,255},
            fillPattern=FillPattern.Forward,
            pattern=LinePattern.None,
            lineColor={238,46,47}),
        Rectangle(
          extent={{-62,2},{-38,-16}},
          lineColor={238,46,47},
            fillColor={255,255,255},
            fillPattern=FillPattern.Solid),
        Rectangle(
          extent={{-14,2},{8,-16}},
          lineColor={238,46,47},
            fillColor={255,255,255},
            fillPattern=FillPattern.Solid),
        Polygon(
          points={{-78,18},{26,18},{10,46},{-66,46},{-78,18}},
            lineColor={238,46,47},
            lineThickness=0.5,
            fillColor={244,125,35},
            fillPattern=FillPattern.Solid),
          Polygon(
            points={{-60,42},{-68,22},{4,22},{6,42},{-60,42}},
            lineThickness=0.5,
            fillColor={28,108,200},
            fillPattern=FillPattern.Forward,
            pattern=LinePattern.None),
          Rectangle(
            extent={{26,0},{40,-40}},
            lineColor={0,0,0},
            pattern=LinePattern.None,
            lineThickness=0.5,
            fillColor={255,255,255},
            fillPattern=FillPattern.Solid)}),                  Diagram(
coordinateSystem(preserveAspectRatio=false)));
{% endraw %}
end building;
"""
        template = ENVIRONMENT.from_string(template_)
        return template


def containers_factory() -> Containers:
    return Containers.load_from_config()



from functools import partial
from pathlib import Path
from typing import Optional, Callable, Any, Dict, List, TYPE_CHECKING

from jinja2 import Environment, FileSystemLoader
from pydantic import BaseModel, Field, ConfigDict, field_validator

from trano.elements.common_base import BaseParameter
from trano.elements.connection import Port
from trano.elements.data_bus.controller_bus import ControllerBus
from trano.elements.figure import Figure
from trano.elements.library import parameters
from trano.elements.library.parameters import default_parameters
from trano.elements.types import BaseVariant, DynamicTemplateCategories
from trano.elements.utils import compose_func, json_component_name

if TYPE_CHECKING:
    from trano.elements.base import BaseElement


class DynamicComponentTemplate(BaseModel):

    template: str
    category: Optional[DynamicTemplateCategories] = None
    function: Callable[[Any], Any] = Field(default=lambda _: {})
    bus: ControllerBus

    def _has_required_attributes(self, element: "BaseElement") -> None:
        for target in self.bus.main_targets():
            if "element" not in target:
                raise ValueError(
                    f"Target {target} should start with the word 'element'"
                )
            attributes = target.split(".")[1:]
            for attr in attributes:
                if not hasattr(element, attr):
                    raise ValueError(
                        f"Element {element} does not have attribute {attr}"
                    )
                element = getattr(element, attr)

    def render(
        self, package_name: str, element: "BaseElement", parameters: Dict[str, Any]
    ) -> str:
        ports = list(self.bus.bus_ports(element))
        environment = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=FileSystemLoader(
                str(Path(__file__).parents[2].joinpath("templates"))
            ),
            autoescape=True,
        )
        environment.filters["enumerate"] = enumerate
        rtemplate = environment.from_string(
            "{% import 'macros.jinja2' as macros %}" + self.template
        )
        component = rtemplate.render(
            element=element,
            package_name=package_name,
            bus_template=self.bus.template,
            bus_ports="\n".join(ports),
            parameters=parameters,
            **self.function(element),
        )

        return component


class LibraryData(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    template: str = ""
    component_template: Optional[DynamicComponentTemplate] = None
    variant: str = BaseVariant.default
    figures: List[Figure] = Field(default=[])
    ports: Callable[[], List[Port]]
    parameter_processing: Callable[[BaseParameter], Dict[str, Any]] = default_parameters
    library: str
    classes: List[str]

    @field_validator("parameter_processing", mode="before")
    @classmethod
    def _parameters_processing_validator(
        cls, value: Dict[str, Any]
    ) -> Callable[[BaseParameter], Dict[str, Any]]:

        if value.get("parameter"):
            function_name = value["function"]
            parameter_processing = partial(
                getattr(parameters, function_name),
                **{function_name: value.get("parameter", {})},
            )
        else:
            parameter_processing = getattr(parameters, value["function"])
        return parameter_processing

    @field_validator("ports", mode="before")
    @classmethod
    def _ports_factory_validator(
        cls, value: List[Dict[str, Any]]
    ) -> Callable[[], List[Port]]:
        return compose_func([Port(**port) for port in value])

    def json_file_name(self) -> str:
        return json_component_name(self.classes)



from trano.elements.base import BaseElement
from trano.elements.types import ContainerTypes


class BaseBoundary(BaseElement): ...


class Boundary(BaseBoundary):
    container_type: ContainerTypes = "ventilation"



import logging
from pathlib import Path
from typing import TYPE_CHECKING, Any, List, Optional, Tuple, Type

from jinja2 import Environment, FileSystemLoader
from pydantic import BaseModel, Field, field_validator, model_validator, computed_field

from trano import elements
from trano.elements.common_base import BaseElementPosition, BasePosition
from trano.elements.types import (
    ConnectionView,
    Flow,
    Medium,
    ContainerTypes,
)
from trano.exceptions import IncompatiblePortsError, ConnectionLimitReached

if TYPE_CHECKING:
    from trano.elements import BaseElement
    from trano.elements.base import ElementPort

INCOMPATIBLE_PORTS = [sorted(["dataBus", "y"])]

logger = logging.getLogger(__name__)


class Port(BaseModel):

    names: list[str]
    targets: List[Any] = Field(default_factory=list)
    expected_ports: List[str] = Field(default_factory=list)
    connected: bool = False
    flow: Flow
    medium: Medium
    multi_connection: bool = False
    multi_object: bool = False
    bus_connection: bool = False
    use_counter: bool = True
    ignore_direction: bool = False
    counter: int = Field(default=1)
    connection_counter: int = Field(default=0)

    def set_connected(self) -> None:
        self.connected = True
        self.connection_counter += 1

    def get_compatible_port(self, port: "Port") -> bool:  # noqa: PLR0911, C901, PLR0912
        if self.medium == Medium.fluid:
            if self.expected_ports:
                return bool(set(port.names).intersection(set(self.expected_ports)))
            else:
                if self.flow == Flow.inlet:
                    return (port.flow in [Flow.outlet]) or (
                        port.flow in [Flow.inlet_or_outlet]
                        and port.multi_connection
                        and port.use_counter
                    )
                if self.flow == Flow.outlet:
                    return port.flow in [Flow.inlet] or (
                        port.flow in [Flow.inlet_or_outlet]
                        and port.multi_connection
                        and port.use_counter
                    )
                if (
                    self.flow == Flow.inlet_or_outlet
                    and self.multi_connection
                    and self.use_counter
                ):
                    return port.flow in [Flow.inlet, Flow.outlet]
                if self.flow == Flow.inlet_or_outlet:
                    return port.flow in [Flow.inlet_or_outlet]

        if self.medium == Medium.data:
            if self.expected_ports:
                return bool(set(port.names).intersection(set(self.expected_ports)))
            elif self.flow == Flow.inlet:
                return port.flow in [Flow.outlet]
            elif self.flow == Flow.outlet:
                return port.flow in [Flow.inlet]
            elif self.flow == Flow.undirected:
                return port.flow in [Flow.undirected, Flow.interchangeable_port]

            else:
                return True
        return self.flow == port.flow

    def similar_flow(self, port: "Port") -> bool:
        return self.flow == port.flow

    def is_complementary(self, port: "Port") -> bool:
        return self.medium == port.medium and self.get_compatible_port(port)

    def is_inlet(self) -> bool:
        return self.flow in [Flow.inlet] and self.medium == Medium.fluid

    def is_outlet(self) -> bool:
        return self.flow in [Flow.outlet] and self.medium == Medium.fluid

    def is_extended_inlet(self) -> bool:
        return (
            self.flow in [Flow.inlet_or_outlet]
            and self.medium == Medium.fluid
            and self.multi_connection
            and self.use_counter
        )

    def is_extended_outlet(self) -> bool:
        return (
            self.flow in [Flow.inlet_or_outlet]
            and self.medium == Medium.fluid
            and self.multi_connection
            and self.use_counter
        )

    def with_directed_flow(self) -> bool:
        return self.flow in [Flow.inlet, Flow.outlet]

    def bidirectional_flow(self) -> bool:
        return self.flow in [Flow.inlet_or_outlet] and self.medium == Medium.fluid

    def without_targets(self) -> "Port":
        return Port.model_validate(self.model_dump(exclude={"targets"}))

    def disconnect(self) -> "Port":
        return Port.model_validate(self.model_dump(exclude={"connected"}))

    def set_ignore_direction(self) -> "Port":
        port = Port.model_validate(self.model_dump())
        port.ignore_direction = True
        return port

    def reset_counter(self) -> "Port":
        self.counter = 1
        return self

    def substract_counter(self) -> "Port":
        if self.counter != 1:
            return Port.model_validate(
                self.model_dump(exclude={"counter"}) | {"counter": self.counter - 1}
            )
        return self

    @field_validator("targets")
    @classmethod
    def validate_targets(cls, values: List[str]) -> List[Type["BaseElement"]]:
        from trano.elements.base import BaseElement

        targets: List[Type[BaseElement]] = []
        for value in values:
            if isinstance(value, str):
                if hasattr(elements, value):
                    targets.append(getattr(elements, value))
                else:
                    raise ValueError(f"Target {value} not found")
            else:
                targets.append(value)
        return targets

    def is_available(self, available_ports: List["Port"]) -> bool:
        if self.multi_connection and self.connected:
            if self.medium == Medium.fluid:
                return not any(
                    (p.medium == self.medium and self.similar_flow(p))
                    for p in available_ports
                )
            else:
                return True
        else:
            return not self.connected

    def is_controllable(self) -> bool:
        from trano.elements.base import Control

        return self.targets is not None and any(
            target == Control for target in self.targets
        )

    def base_equation(
        self,
        merged_number: int,
        element_name: Optional[str] = None,
        element_position: Optional[BasePosition] = None,
    ) -> List["BasePartialConnection"]:
        element_position = element_position or BasePosition()
        partial_connections = []
        for sub_port_number, name in enumerate(self.names):
            if self.multi_connection and self.bus_connection:
                first_counter = self.counter
                last_counter = self.counter + merged_number - 1
                counter = (
                    f"{first_counter}"
                    if first_counter == last_counter
                    else f"{first_counter}:{last_counter}"
                )
                if element_name:
                    equation = (
                        f"{element_name}[{counter}].{name}"
                        if self.multi_object
                        else f"{element_name}.{name}[{counter}]"
                    )
                else:
                    equation = f"{name}[{counter}]" if self.multi_object else f"{name}"
                self.counter = last_counter + 1
            elif self.multi_connection and self.use_counter:
                equation = (
                    f"{element_name}.{name}[{self.counter}]"
                    if element_name
                    else f"{name}[{self.counter}]"
                )
                self.counter += 1
            elif element_name:
                equation = f"{element_name}.{name}"
            else:
                equation = f"{name}"
            partial_connections.append(
                BasePartialConnection(
                    equation=equation,
                    position=element_position,
                    port=self,
                    sub_port=sub_port_number,
                    name=element_name,
                )
            )

        return partial_connections

    def link(  # noqa: PLR0913
        self,
        merged_number: int,
        element_name: str | None,
        element_position: BasePosition,
        container_type: ContainerTypes | None,
        connected_container_type: ContainerTypes | None,
    ) -> list["PartialConnection"]:
        base_equations = self.base_equation(
            merged_number, element_name, element_position
        )

        return [
            PartialConnection(
                **(
                    b.model_dump()
                    | {
                        "connected_container_type": connected_container_type,
                        "container_type": container_type,
                    }
                )
            )
            for b in base_equations
        ]


def connection_color(edge: Tuple["ElementPort", "ElementPort"]) -> ConnectionView:
    from trano.elements.bus import DataBus
    from trano.elements.envelope import BaseSimpleWall
    from trano.elements.system import Weather, System
    from trano.elements.base import Control

    if any(isinstance(e, BaseSimpleWall) for e in edge):
        return ConnectionView(color="{191,0,0}", thickness=0.1)
    if any(isinstance(e, DataBus) for e in edge):
        return ConnectionView(color=None, thickness=0.05)
    if any(isinstance(e, Weather) for e in edge):
        return ConnectionView(color=None, thickness=0.05)
    if any(isinstance(e, Control) for e in edge):
        return ConnectionView(color="{139, 0, 0}", thickness=0.1, pattern="Dash")
    if all(isinstance(e, System) for e in edge):
        return ConnectionView(color="{0, 0, 139}", thickness=0.75)
    return ConnectionView()


def check_flow_direction(first_port: Port, second_port: Port) -> bool:
    if first_port.medium == Medium.fluid and second_port.medium == Medium.fluid:
        if first_port.ignore_direction and second_port.ignore_direction:
            return first_port.get_compatible_port(second_port)
        elif second_port.with_directed_flow() and first_port.with_directed_flow():
            return first_port.is_outlet() and second_port.is_inlet()
        else:
            return (
                second_port.bidirectional_flow()
                and first_port.bidirectional_flow()
                or (
                    (first_port.is_outlet() and second_port.is_extended_inlet())
                    or (first_port.is_extended_outlet() and second_port.is_inlet())
                )
            )
    else:
        return True


def connect(
    edge_first: "ElementPort", edge_second: "ElementPort"
) -> list["Connection"]:
    connections = []
    try:
        for first_port in edge_first.ports:
            for second_port in edge_second.ports:
                available = first_port.is_available(
                    edge_first.available_ports()
                ) and second_port.is_available(edge_second.available_ports())
                has_targets = edge_second.has_target(
                    first_port.targets
                ) and edge_first.has_target(second_port.targets)
                complementarity = first_port.is_complementary(second_port)
                flow_direction = check_flow_direction(first_port, second_port)
                if available and has_targets and complementarity and flow_direction:
                    merged_number = max(
                        edge_first.merged_number, edge_second.merged_number
                    )
                    left_right = list(
                        zip(
                            first_port.link(
                                merged_number,
                                edge_first.name,
                                edge_first.position,
                                edge_first.container_type,
                                edge_second.container_type,
                            ),
                            second_port.link(
                                merged_number,
                                edge_second.name,
                                edge_second.position,
                                edge_second.container_type,
                                edge_first.container_type,
                            ),
                            strict=True,
                        )
                    )
                    for left, right in left_right:
                        connection = Connection(
                            left=left,
                            right=right,
                            connection_view=connection_color((edge_first, edge_second)),
                        )
                        connections.append(connection)
                    first_port.set_connected()
                    second_port.set_connected()
                    current_connection_numbers = len(connections)
                    allowed_connections = edge_first.get_connection_per_target(
                        edge_second.element_type
                    )
                    if current_connection_numbers >= allowed_connections:
                        raise ConnectionLimitReached
    except ConnectionLimitReached:
        ...

    return connections


class BasePartialConnection(BaseElementPosition):
    name: Optional[str] = None
    equation: str
    port: Port
    sub_port: int


class BasePartialConnectionWithContainerType(BasePartialConnection):
    container_type: Optional[ContainerTypes] = None


class PartialConnection(BasePartialConnectionWithContainerType):
    connected_container_type: Optional[ContainerTypes] = None

    def reset_port_counter(self) -> "PartialConnection":
        self.port.reset_counter()
        return self

    def to_base_partial_connection(self) -> BasePartialConnection:
        return BasePartialConnection.model_validate(
            self.model_dump(exclude={"connected_container_type", "container_type"})
        )


class Connection(BaseModel):
    right: PartialConnection
    left: PartialConnection
    connection_view: ConnectionView = Field(default=ConnectionView())

    def in_the_same_container(self) -> bool:
        return bool(self.left.name is not None) and bool(self.right.name is not None)

    def reset_counters(self) -> "Connection":
        self.left.reset_port_counter()
        self.right.reset_port_counter()
        return self

    def equation_view(self) -> Tuple[str, ...]:
        return tuple(sorted([self.left.equation, self.right.equation]))

    @computed_field
    def equation(self) -> str:
        environment = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=FileSystemLoader(
                str(Path(__file__).parents[1].joinpath("templates"))
            ),
            autoescape=True,
        )
        annotation_template = environment.from_string(
            """{% import 'macros.jinja2' as macros %}        
        connect({{ connection.left.equation }},{{ connection.right.equation }})
            {% if not connection.connection_view.disabled %}
        annotation (Line(
        points={{ macros.connect_path(connection.path) }},
            {% if connection.connection_view.color %}
        color={{ connection.connection_view.color }},
            {% endif %}
        thickness={{ connection.connection_view.thickness }},pattern =
        LinePattern.{{ connection.connection_view.pattern }},
        smooth=Smooth.None))
            {% endif %}
            ;"""
        )

        return annotation_template.render(
            connection=self,
        )

    @model_validator(mode="after")
    def _connection_validator(self) -> "Connection":
        if (
            self.right.position.container.is_empty()
            or self.left.position.container.is_empty()
        ):
            logger.debug(
                f"Connection position still empty for {self.right.name} and {self.left.name}."
            )
        if (
            sorted(
                [
                    part.split(".")[-1]
                    for part in [self.right.equation, self.left.equation]
                ]
            )
            in INCOMPATIBLE_PORTS
        ):
            raise IncompatiblePortsError(
                f"Incompatible ports {self.right.equation} and {self.left.equation}."
            )
        return self

    @property
    def path(self) -> List[List[float] | Tuple[float, float]]:
        if (
            self.left.position.global_.location.c_.x
            < self.right.position.global_.location.c_.x
        ):
            mid_path = (
                self.right.position.global_.location.c_.x
                - self.left.position.global_.location.c_.x
            ) / 2
            return [
                self.left.position.global_.coordinate(),
                (
                    self.left.position.global_.location.c_.x + mid_path,
                    self.left.position.global_.location.c_.y,
                ),
                (
                    self.right.position.global_.location.c_.x - mid_path,
                    self.right.position.global_.location.c_.y,
                ),
                self.right.position.global_.coordinate(),
            ]

        else:
            mid_path = (
                self.left.position.global_.location.c_.x
                - self.right.position.global_.location.c_.x
            ) / 2
            return [
                self.left.position.global_.coordinate(),
                (
                    self.left.position.global_.location.c_.x - mid_path,
                    self.left.position.global_.location.c_.y,
                ),
                (
                    self.right.position.global_.location.c_.x + mid_path,
                    self.right.position.global_.location.c_.y,
                ),
                self.right.position.global_.coordinate(),
            ]


class ContainerConnection(Connection):
    source: Tuple[str, str]

    def get_container_equation(
        self,
    ) -> Optional[BasePartialConnectionWithContainerType]:
        if self.right.name is None:
            return self.right
        if self.left.name is None:
            return self.left
        return None

    @property
    def path(self) -> List[List[float] | Tuple[float, float]]:
        if (
            self.left.position.container.location.c_.x
            < self.right.position.container.location.c_.x
        ):
            mid_path = (
                self.right.position.container.location.c_.x
                - self.left.position.container.location.c_.x
            ) / 2
            return [
                self.left.position.container.coordinate(),
                (
                    self.left.position.container.location.c_.x + mid_path,
                    self.left.position.container.location.c_.y,
                ),
                (
                    self.right.position.container.location.c_.x - mid_path,
                    self.right.position.container.location.c_.y,
                ),
                self.right.position.container.coordinate(),
            ]

        else:
            mid_path = (
                self.left.position.container.location.c_.x
                - self.right.position.container.location.c_.x
            ) / 2
            return [
                self.left.position.container.coordinate(),
                (
                    self.left.position.container.location.c_.x - mid_path,
                    self.left.position.container.location.c_.y,
                ),
                (
                    self.right.position.container.location.c_.x + mid_path,
                    self.right.position.container.location.c_.y,
                ),
                self.right.position.container.coordinate(),
            ]



import itertools
import logging
import shutil
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Type, cast

import networkx as nx
from jinja2 import Environment, FileSystemLoader
from networkx import DiGraph

from trano.elements import (
    BaseElement,
    Connection,
    Control,
    DynamicTemplateCategories,
    InternalElement,
    connect,
)
from trano.elements.base import ElementPort
from trano.elements.bus import DataBus
from trano.elements.construction import (
    extract_properties,
    Construction,
    Layer,
    Material,
)
from trano.elements.containers import containers_factory, ContainerInput
from trano.elements.library.library import Library
from trano.elements.space import Space
from trano.elements.system import (
    System,
    TemperatureSensor,
    ThreeWayValve,
    Weather,
)
from trano.elements.types import Tilt
from trano.elements.utils import generate_normalized_layout
from trano.exceptions import SystemsNotConnectedError

logger = logging.getLogger(__name__)


def all_subclasses(cls: Type[BaseElement]) -> List[Type[BaseElement]]:
    return cls.__subclasses__() + [
        g for s in cls.__subclasses__() for g in all_subclasses(s)
    ]


def reset_element_names() -> None:
    from trano.data_models.conversion import COUNTER

    COUNTER.clear()
    for subclass in all_subclasses(BaseElement):
        subclass.name_counter = 0


def default_internal_wall_construction() -> Construction:
    return Construction(
        name="internal_wall",
        layers=[
            Layer(
                material=Material(
                    name="brick",
                    thermal_conductivity=0.89,
                    density=1920,
                    specific_heat_capacity=790,
                ),
                thickness=0.2,
            ),
        ],
    )


class Network:  # : PLR0904, #TODO: fix this
    def __init__(
        self,
        name: str,
        library: Optional[Library] = None,
        external_data: Optional[Path] = None,
        diagram_scale: Optional[int] = None,
    ) -> None:
        reset_element_names()
        self.graph: DiGraph = DiGraph()
        self.edge_attributes: List[Connection] = []
        self.name: str = name
        self._system_controls: List[Control] = []
        self.library = library or Library.load_default()
        self.external_data = external_data
        self.containers = containers_factory()
        self.diagram_scale = diagram_scale or 1000
        self.dynamic_components: Dict[DynamicTemplateCategories, List[str]] = {
            "ventilation": [],
            "control": [],
            "boiler": [],
        }

    @property
    def diagram_size(self) -> str:
        return f"{{{{{-50},{-50}}},{{{self.diagram_scale},{self.diagram_scale}}}}}"

    def get_node(self, element: Type[BaseElement]) -> Optional[BaseElement]:
        element_node = [node for node in self.graph.nodes if isinstance(node, element)]
        if element_node:
            return element_node[0]
        return None

    def get_edge(
        self, first_edge: Type[BaseElement], second_edge: Type[BaseElement]
    ) -> Connection:
        return cast(
            Connection,
            next(
                edge
                for edge in self.graph.edges
                if (
                    (
                        isinstance(edge[0], first_edge)
                        and isinstance(edge[1], second_edge)
                    )
                    or (
                        isinstance(edge[1], first_edge)
                        and isinstance(edge[0], second_edge)
                    )
                )
            ),
        )

    def add_node(self, node: BaseElement) -> None:

        if not node.libraries_data:
            return
            # TODO: check better option here!!
        found_library = node.assign_library_property(self.library)
        if not found_library:
            return

        if node not in self.graph.nodes:
            self.graph.add_node(node)
        if (isinstance(node, System) and node.control) and (
            node.control not in self.graph.nodes
        ):
            node_control = node.control
            if not node_control.libraries_data:
                raise Exception(
                    f"No library data defined for NOde of type {type(node).__name__}"
                )
            node_control.assign_library_property(self.library)
            self.graph.add_node(node_control)
            self.graph.add_edge(node, node_control)
            node_control.controllable_element = node

    def _add_subsequent_systems(self, systems: List[System]) -> None:
        for system1, system2 in zip(systems[:-1], systems[1:]):
            if not self.graph.has_node(system1):  # type: ignore
                self.add_node(system1)
            if not self.graph.has_node(system2):  # type: ignore
                self.add_node(system2)
            self.graph.add_edge(
                system1,
                system2,
            )

    def connect_spaces(
        self,
        space_1: "Space",
        space_2: "Space",
        internal_element: Optional[
            "InternalElement"
        ] = None,  # TODO: this should not be optional
    ) -> None:
        internal_element = internal_element or InternalElement(
            name=f"internal_{space_1.name}_{space_2.name}",
            surface=10,
            azimuth=10,
            construction=default_internal_wall_construction(),
            tilt=Tilt.wall,
        )
        if space_1.position.is_global_empty() or space_2.position.is_global_empty():
            raise Exception("Position not assigned to spaces")
        internal_element.position.between_two_objects(
            space_1.position.global_, space_2.position.global_
        )

        self.add_node(internal_element)
        self.graph.add_edge(
            space_1,
            internal_element,
        )
        self.graph.add_edge(
            space_2,
            internal_element,
        )
        space_1.internal_elements.append(internal_element)
        space_2.internal_elements.append(internal_element)

    def connect_system(self, space: "Space", system: "System") -> None:
        self.graph.add_edge(
            space,
            system,
        )

    def connect_elements(self, element_1: BaseElement, element_2: BaseElement) -> None:
        for element in [element_1, element_2]:
            if element not in self.graph.nodes:
                self.add_node(element)
        self.graph.add_edge(element_1, element_2)

    def connect_systems(self, system_1: System, system_2: System) -> None:

        if system_1 not in self.graph.nodes:
            self.add_node(system_1)
            if system_1.control:
                if system_1.control not in self.graph.nodes:
                    self.add_node(system_1.control)
                    self._system_controls.append(system_1.control)
                self.graph.add_edge(system_1, system_1.control)
                # TODO: check if it is controllable the system

        if system_2 not in self.graph.nodes:
            self.add_node(system_2)
            if system_2.control:
                if system_2.control not in self.graph.nodes:
                    self.add_node(system_2.control)
                    self._system_controls.append(system_2.control)
                self.graph.add_edge(system_2, system_2.control)
        if (
            isinstance(system_2, ThreeWayValve)
            and isinstance(system_1, TemperatureSensor)
        ) or (
            isinstance(system_1, ThreeWayValve)
            and isinstance(system_2, TemperatureSensor)
        ):
            if system_2.control:
                self.graph.add_edge(system_2.control, system_1)
            if system_1.control:
                self.graph.add_edge(system_1.control, system_2)
        self.graph.add_edge(system_1, system_2)

    def connect_edges(
        self, edge: Tuple[BaseElement, BaseElement]  # :  PLR6301
    ) -> list[Connection]:
        connection = connect(
            ElementPort.from_element(edge[0]), ElementPort.from_element(edge[1])
        )
        if not connection:
            logger.warning(
                f"Connection not possible between {edge[0].name} and {edge[1].name}"
            )
        return connection

    def merge_spaces(self, space_1: "Space", space_2: "Space") -> None:
        internal_elements = nx.shortest_path(self.graph, space_1, space_2)[1:-1]
        merged_space = space_1 + space_2
        merged_space.internal_elements = internal_elements
        self.graph = nx.contracted_nodes(self.graph, merged_space, space_2)

    def assign_nodes_position(self) -> None:
        global_position = generate_normalized_layout(self, scale=self.diagram_scale)
        container_positions = {
            container.name: generate_normalized_layout(
                self,
                scale=container.layout.scale,
                origin=container.layout.bottom_left,
                container_type=container.name,
            )
            for container in self.containers.containers
        }
        for node in self.graph.nodes:
            node.set_position(global_position)
            if node.container_type and container_positions[node.container_type]:
                node.set_position(
                    container_positions[node.container_type], global_=False
                )

    def connect(self) -> None:
        self.assign_nodes_position()
        data_buses = [bus for bus in list(self.graph.nodes) if isinstance(bus, DataBus)]

        data_bus = data_buses[0] if data_buses else None
        new_edges = [edge for edge in self.graph.edges if data_bus not in edge]
        edge_with_databus = [edge for edge in self.graph.edges if data_bus in edge]
        edges_with_bus_without_space = [
            edge
            for edge in edge_with_databus
            if not any(isinstance(e, Space) for e in edge)
        ]
        edges_with_bus_with_space = sorted(
            [
                edge
                for edge in edge_with_databus
                if any(isinstance(e, Space) for e in edge)
            ],
            key=lambda e_: next(e for e in e_ if isinstance(e, Space)).name,
        )
        # Sorting is necessary here since we need to keep the
        # same index for the same space indatabus
        # TODO: not sure where to put this!!!!
        for edge in (
            new_edges + edges_with_bus_without_space + edges_with_bus_with_space
        ):
            self.edge_attributes += self.connect_edges(edge)

    def set_weather_path_to_container_path(self, project_path: Path) -> None:
        for node in self.graph.nodes:
            if (
                isinstance(node, Weather)
                and hasattr(node.parameters, "path")
                and node.parameters.path is not None  # type: ignore
            ):
                # TODO: type ognore needs to be fixed
                old_path = Path(node.parameters.path).resolve()  # type: ignore
                if not old_path.exists():
                    parents = [Path.cwd(), *Path.cwd().parents]
                    for parent in parents:
                        old_path = next(parent.rglob(old_path.name), None)  # type: ignore
                        if old_path and old_path.exists():
                            break
                    if not old_path or not old_path.exists():
                        raise FileNotFoundError(f"File {old_path} not found")
                new_path = project_path.joinpath(old_path.name)
                shutil.copy(old_path, new_path)
                # TODO: this is not correct
                node.parameters.path = f'"/simulation/{old_path.name}"'  # type: ignore

    def model(
        self, include_container: bool = True, data_bus: Optional[DataBus] = None
    ) -> str:
        Space.counter = 0
        for node in self.graph.nodes:
            node.assign_container_type(self)
            node.processing(self)
        if not self.get_node(DataBus):
            data_bus = data_bus or DataBus()
            data_bus.add_to_network(self)
        for node in self.graph.nodes:
            node.configure(self)
        self.connect()
        data = extract_properties(self.library, self.name, self.graph.nodes)
        component_models = []
        for node in self.graph.nodes:
            model = node.model(self)
            if model:
                component_models.append(model)

        container_input = ContainerInput(
            nodes=list(self.graph.nodes),
            connections=self.edge_attributes,
            data=data,
            medium=self.library.medium,
        )
        container_model = self.containers.build(container_input)
        element_models = [c.model for c in component_models]
        environment = Environment(
            trim_blocks=True,
            lstrip_blocks=True,
            loader=FileSystemLoader(str(Path(__file__).parent.joinpath("templates"))),
            autoescape=True,
        )
        environment.filters["frozenset"] = frozenset
        environment.filters["enumerate"] = enumerate
        template = environment.get_template("base.jinja2")
        if not all(n.system_ports_connected() for n in self.graph.nodes):
            raise SystemsNotConnectedError(
                f"""Not all system ports are connected. 
            The following are not connected: {[nm for n in self.graph.nodes for nm in n.non_connected_ports_name() 
                                               if not n.system_ports_connected() and nm]}"""
            )
        return template.render(
            network=self,
            data=data,
            element_models=element_models,
            library=self.library,
            databus=data_bus,
            dynamic_components=self.dynamic_components,
            diagram_size=self.diagram_size,
            containers=container_model if include_container else [],
            main=self.containers.main if include_container else "",
            include_container=include_container,
        )

    def add_boiler_plate_spaces(
        self,
        spaces: list[Space],
        create_internal: bool = True,
        weather: Optional[Weather] = None,
    ) -> None:
        for space in spaces:
            space.add_to_network(self)
        if create_internal:
            for combination in itertools.combinations(spaces, 2):
                self.connect_spaces(*combination)
        weather = weather or Weather()
        weather.position.set_global(-100, 200)
        self.add_node(weather)
        for space in spaces:
            self.connect_system(space, weather)



import math
from math import sqrt
from typing import TYPE_CHECKING, Dict, List, Optional, Type, Union

import numpy as np
from pydantic import BaseModel, model_validator, computed_field, Field

from trano.elements.base import BaseElement
from trano.elements.construction import Construction, Glass
from trano.elements.types import Azimuth, Tilt, ContainerTypes, TILT_MAPPING

if TYPE_CHECKING:
    pass


class BaseWall(BaseElement):
    container_type: ContainerTypes = "envelope"

    # @computed_field  # type: ignore
    @property
    def length(self) -> int:
        if hasattr(self, "surfaces"):
            return len(self.surfaces)
        return 1


class BaseSimpleWall(BaseWall):
    surface: float | int
    azimuth: float | int
    tilt: Tilt
    construction: Construction | Glass


class BaseInternalElement(BaseSimpleWall): ...


class BaseFloorOnGround(BaseSimpleWall): ...


class BaseExternalWall(BaseSimpleWall): ...


class BaseWindow(BaseSimpleWall):
    width: Optional[float] = None
    height: Optional[float] = None

    @model_validator(mode="after")
    def width_validator(self) -> "BaseWindow":
        if self.width is None and self.height is None:
            self.width = sqrt(self.surface)
            self.height = sqrt(self.surface)
        elif self.width is not None and self.height is None:
            self.height = self.surface / self.width
        elif self.width is None and self.height is not None:
            self.width = self.surface / self.height
        else:
            ...
        return self


def _get_element(
    construction_type: str,
    base_walls: list[BaseExternalWall | BaseWindow | BaseFloorOnGround],
    construction: Construction | Glass,
) -> List[Union[BaseExternalWall | BaseWindow | BaseFloorOnGround]]:
    return [
        getattr(base_wall, construction_type)
        for base_wall in base_walls
        if base_wall.construction == construction
    ]


class MergedBaseWall(BaseWall):
    surfaces: List[float | int]
    azimuths: List[float | int]
    tilts: List[Tilt]
    constructions: List[Construction | Glass]

    @classmethod
    def from_base_elements(
        cls, base_walls: list[BaseExternalWall | BaseWindow | BaseFloorOnGround]
    ) -> List["MergedBaseWall"]:
        merged_walls = []
        unique_constructions = {base_wall.construction for base_wall in base_walls}

        for construction in unique_constructions:
            data: Dict[
                str,
                list[BaseExternalWall | BaseWindow | BaseFloorOnGround],
            ] = {
                "azimuth": [],
                "tilt": [],
                "name": [],
                "surface": [],
            }
            for construction_type in data:
                data[construction_type] = _get_element(
                    construction_type, base_walls, construction
                )
            merged_wall = cls(
                name=f"merged_{'_'.join(data['name'])}",  # type: ignore
                surfaces=data["surface"],
                azimuths=data["azimuth"],
                tilts=data["tilt"],
                constructions=[construction],
            )
            merged_walls.append(merged_wall)
        return sorted(merged_walls, key=lambda x: x.name)  # type: ignore #TODO: what is the issue with this!!!


class MergedBaseWindow(MergedBaseWall): ...


class MergedBaseExternalWall(MergedBaseWall): ...


class ExternalDoor(BaseExternalWall): ...


class ExternalWall(ExternalDoor): ...


class FloorOnGround(BaseFloorOnGround):
    azimuth: float | int = Azimuth.south
    tilt: Tilt = Tilt.floor


class InternalElement(BaseInternalElement): ...


class MergedFloor(MergedBaseWall): ...


class MergedExternalWall(MergedBaseExternalWall): ...


class MergedWindows(MergedBaseWindow):
    widths: List[float | int]
    heights: List[float | int]

    @classmethod
    def from_base_windows(cls, base_walls: List["BaseWindow"]) -> List["MergedWindows"]:
        merged_windows = []
        unique_constructions = {base_wall.construction for base_wall in base_walls}

        for construction in unique_constructions:
            data: Dict[
                str, List[Union["ExternalWall", "FloorOnGround", "BaseWindow", str]]
            ] = {
                "azimuth": [],
                "tilt": [],
                "name": [],
                "surface": [],
                "width": [],
                "height": [],
            }
            for construction_type in data:
                data[construction_type] = _get_element(
                    construction_type, base_walls, construction  # type: ignore
                )
            merged_window = cls(
                name=f"merged_{'_'.join(data['name'])}",  # type: ignore
                surfaces=data["surface"],
                azimuths=data["azimuth"],
                tilts=data["tilt"],
                constructions=[construction],
                heights=data["height"],
                widths=data["width"],
            )
            merged_windows.append(merged_window)
        return sorted(merged_windows, key=lambda x: x.name)  # type: ignore


class Window(BaseWindow): ...


class WindowedWall(BaseSimpleWall): ...


def parallel_resistance(resistances: list[float]) -> float:
    return 1 / sum([1 / resistance for resistance in resistances])


class WallParameters(BaseModel):
    number: int
    surfaces: list[float]
    azimuths: list[float]
    u_values: list[float]
    layers: list[str]
    tilts: list[Tilt]
    window_area_by_orientation: list[float]
    type: str
    average_resistance_external: float = Field(default=0)
    average_resistance_external_remaining: float = Field(default=0)
    total_thermal_capacitance: float = Field(default=0)
    total_thermal_resistance: float = Field(default=0)

    def azimuths_to_radians(self) -> list[float]:
        return [math.radians(azimuth) for azimuth in self.azimuths]

    def tilts_to_radians(self) -> list[float]:
        return [math.radians(TILT_MAPPING[tilt.value]) for tilt in self.tilts]

    @computed_field
    def average_u_value(self) -> float:
        if not self.u_values:
            return 0
        return sum(self.u_values) / len(self.u_values)

    @classmethod
    def from_neighbors(
        cls,
        neighbors: list["BaseElement"],
        wall: Type["BaseSimpleWall"],
        filter: Optional[list[str]] = None,
        suffix_type: Optional[str] = None,
    ) -> "WallParameters":
        constructions = [
            neighbor
            for neighbor in neighbors
            if isinstance(neighbor, wall)
            if neighbor.name not in (filter or [])
        ]
        window_area_by_orientation = []
        number = len(constructions)
        surfaces = [
            exterior_construction.surface for exterior_construction in constructions
        ]
        total_surface = sum(surfaces)
        azimuths = [
            exterior_construction.azimuth for exterior_construction in constructions
        ]
        layers = [
            exterior_construction.construction.name
            for exterior_construction in constructions
        ]
        u_values = [
            exterior_construction.construction.u_value
            for exterior_construction in constructions
        ]
        average_resistance_external = np.mean(
            [  # type: ignore
                exterior_construction.construction.resistance_external
                for exterior_construction in constructions
            ]
        )
        average_resistance_external_remaining = np.mean(
            [  # type: ignore
                exterior_construction.construction.resistance_external_remaining
                for exterior_construction in constructions
            ]
        )
        total_thermal_capacitance = total_surface * np.mean(
            [  # type: ignore
                exterior_construction.construction.total_thermal_capacitance
                for exterior_construction in constructions
            ]
        )
        total_thermal_resistance = total_surface * np.mean(
            [
                exterior_construction.construction.total_thermal_resistance
                for exterior_construction in constructions
            ]
        )

        tilt = [exterior_construction.tilt for exterior_construction in constructions]
        type = wall.__name__ if not suffix_type else f"{wall.__name__}{suffix_type}"
        if issubclass(wall, BaseWindow):
            external_walls = [
                neighbor for neighbor in neighbors if isinstance(neighbor, ExternalWall)
            ]
            azimuth_surface = {
                construction.azimuth: construction.surface
                for construction in constructions
            }
            window_area_by_orientation = [
                (azimuth_surface.get(exterior_construction.azimuth, 0))
                for exterior_construction in external_walls
            ]
        return cls(
            number=number,
            surfaces=surfaces,
            azimuths=azimuths,
            layers=layers,
            tilts=tilt,
            type=type,
            u_values=u_values,
            window_area_by_orientation=window_area_by_orientation,
            average_resistance_external=average_resistance_external,
            average_resistance_external_remaining=average_resistance_external_remaining,
            total_thermal_capacitance=total_thermal_capacitance,
            total_thermal_resistance=total_thermal_resistance,
        )


class VerticalWallParameters(WallParameters):
    @classmethod
    def from_neighbors_(
        cls, neighbors: list["BaseElement"], wall: Type["BaseSimpleWall"]
    ) -> "VerticalWallParameters":
        neighbors = [n for n in neighbors if hasattr(n, "tilt") and n.tilt == Tilt.wall]
        return cls.from_neighbors(neighbors, wall, suffix_type="VerticalOnly")  # type: ignore


class RoofWallParameters(WallParameters):
    @classmethod
    def from_neighbors_(
        cls, neighbors: list["BaseElement"], wall: Type["BaseSimpleWall"]
    ) -> "RoofWallParameters":
        neighbors = [
            n for n in neighbors if hasattr(n, "tilt") and n.tilt == Tilt.ceiling
        ]
        return cls.from_neighbors(neighbors, wall, suffix_type="Roof")  # type: ignore


class ExternalWallParameters(WallParameters):
    @classmethod
    def from_neighbors_(
        cls, neighbors: list["BaseElement"], wall: Type["BaseSimpleWall"]
    ) -> "RoofWallParameters":
        return cls.from_neighbors(neighbors, wall, suffix_type="External")  # type: ignore


class WindowedWallParameters(WallParameters):
    window_layers: list[str]
    window_width: list[float]
    window_height: list[float]
    included_external_walls: list[str]

    @classmethod
    def from_neighbors(cls, neighbors: list["BaseElement"]) -> "WindowedWallParameters":  # type: ignore

        windows = [
            neighbor for neighbor in neighbors if isinstance(neighbor, BaseWindow)
        ]
        surfaces = []
        azimuths = []
        layers = []
        tilts = []
        window_layers = []
        window_width = []
        window_height = []
        included_external_walls = []
        u_values = []
        window_area_by_orientation: List[float] = []
        for window in windows:
            wall = get_common_wall_properties(neighbors, window)
            surfaces.append(wall.surface)
            azimuths.append(wall.azimuth)
            layers.append(wall.construction.name)
            tilts.append(wall.tilt)
            window_layers.append(window.construction.name)
            window_width.append(window.width)
            window_height.append(window.height)
            included_external_walls.append(wall.name)
            u_values.append(wall.construction.u_value)
        return cls(
            number=len(windows),
            surfaces=surfaces,
            azimuths=azimuths,
            layers=layers,
            u_values=u_values,
            tilts=tilts,
            type="WindowedWall",
            window_layers=window_layers,
            window_width=window_width,
            window_height=window_height,
            included_external_walls=included_external_walls,
            window_area_by_orientation=window_area_by_orientation,
        )


def get_common_wall_properties(
    neighbors: list["BaseElement"], window: BaseWindow
) -> BaseSimpleWall:
    walls = [
        neighbor
        for neighbor in neighbors
        if isinstance(neighbor, ExternalWall)
        and neighbor.azimuth == window.azimuth
        and Tilt.wall == neighbor.tilt
    ]
    similar_properties = (
        len({w.azimuth for w in walls}) == 1
        and len({w.tilt for w in walls}) == 1
        and len({w.construction.name for w in walls}) == 1
    )

    if not similar_properties:
        raise NotImplementedError
    return BaseSimpleWall(
        surface=sum([w.surface for w in walls]),
        name=walls[0].name,
        tilt=walls[0].tilt,
        azimuth=walls[0].azimuth,
        construction=walls[0].construction,
    )



from pathlib import Path
from typing import List, Any, Dict

import yaml
from pydantic import BaseModel


from trano.elements.library.base import LibraryData


class Components(BaseModel):
    components: List[Dict[str, Any]]

    @classmethod
    def load(cls) -> "Components":
        libraries_json_path = Path(__file__).parent.joinpath("models")
        libraries_json_path.mkdir(exist_ok=True)
        components = []
        for file in libraries_json_path.rglob("*.yaml"):
            file_data = yaml.safe_load(file.read_text())
            components += file_data

        return cls(components=components)

    def get_components(self, component_name: str) -> List[LibraryData]:
        libraries_data = [
            LibraryData.model_validate(c)
            for c in self.components
            if component_name in c["classes"]
        ]
        if len({(ld.variant, ld.library) for ld in libraries_data}) != len(
            libraries_data
        ):
            raise ValueError(f"Duplicate variant for {component_name}")
        return libraries_data


COMPONENTS = Components.load()



import logging
import re
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, List, Tuple, Dict, Optional

import networkx as nx
import yaml

from trano import elements
from trano.elements.common_base import Point

if TYPE_CHECKING:

    from trano.elements import Port
    from trano.topology import Network

logger = logging.getLogger(__name__)


class BlockStyleDumper(yaml.Dumper):
    def represent_scalar(
        self, tag: Any, value: Any, style: Optional[str] = None  # noqa: ANN401
    ) -> Any:  # noqa: ANN401
        if isinstance(value, str) and ("\n" in value):
            style = "|"
        return super().represent_scalar(tag, value, style)


def to_snake_case(name: str) -> str:
    s1 = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", s1).lower()


def compose_func(ports_: List["Port"]) -> Callable[[], List["Port"]]:
    return lambda: ports_


def _get_type(_type: Any) -> Any:  # noqa: ANN401
    if _type == "string":
        return str
    elif _type == "float":
        return float
    elif _type == "integer":
        return int
    elif _type == "boolean":
        return bool
    else:
        raise Exception("Unknown type")


def _get_default(v: Any) -> Any:  # noqa: ANN401
    if "ifabsent" not in v:
        return None
    tag = v["range"]
    if tag == "integer":
        tag = "int"
    value = v["ifabsent"].replace(tag, "")[1:-1]
    if value == "None":
        return None

    try:
        return _get_type(v["range"])(value)
    except Exception as e:

        raise e


# TODO: class names should be standardized!!
def import_element_function(function_name: str) -> Any:  # noqa: ANN401
    attribute = [
        attribute
        for attribute in elements.__all__
        if attribute.lower() == function_name.lower()
    ]
    if len(attribute) > 1:
        raise Exception(f"Element {function_name} has more than one match")
    if len(attribute) == 0:
        raise Exception(f"Element {function_name} not found")
    return getattr(elements, attribute[0])


def generate_normalized_layout(
    network: "Network",
    scale: int = 200,
    origin: Optional[Point] = None,
    container_type: Optional[str] = None,
) -> Dict[str, Tuple[float, float]]:
    def normalize(
        value: float, min_val: float, max_val: float, scale: int = 200
    ) -> float:
        return (
            ((value - min_val) / (max_val - min_val)) * scale
            if max_val != min_val
            else scale / 2
        )

    origin = origin or Point(x=0, y=0)
    new_graph = nx.DiGraph()
    edges = [
        (e[0].name, e[1].name)
        for e in network.graph.edges
        if (
            e[0].container_type == container_type
            and e[1].container_type == container_type
        )
        or (container_type is None)
    ]
    nodes = [
        n.name
        for n in network.graph.nodes
        if (n.container_type == container_type) or (container_type is None)
    ]
    if not nodes:
        return {}
    new_graph.add_nodes_from(nodes)
    new_graph.add_edges_from(edges)
    try:
        pos = nx.nx_pydot.pydot_layout(new_graph, prog="sfdp")
    except Exception as e:
        logger.error(
            f"Error generating layout using graphviz. {e}. Graphviz is probably not installed."
        )
        pos = nx.random_layout(new_graph)
    x_values, y_values = zip(*pos.values())
    x_min, x_max = min(x_values), max(x_values)
    y_min, y_max = min(y_values), max(y_values)
    return {
        node: (
            origin.c_.x + normalize(x, x_min, x_max, scale),
            origin.c_.y + normalize(y, y_min, y_max, scale),
        )
        for node, (x, y) in pos.items()
    }


def wrap_with_raw(template: str) -> str:
    pattern = (
        r"(\{\s*-?\d+\s*,\s*-?\d+\s*(?:,\s*-?\d+\s*)?\})|"
        r"(\{\{\s*-?\d+\s*,\s*-?\d+\s*\},\{\s*-?\d+\s*,\s*-?\d+\s*\}\})"
    )
    return re.sub(
        pattern, lambda m: f"{{% raw %}} {m.group(0)} {{% endraw %}}", template
    )


def camel_to_snake(camel_str: str) -> str:
    snake_str = re.sub(r"(?<!^)(?=[A-Z])", "_", camel_str).lower()
    return snake_str


def json_component_name(classes: List[str]) -> str:
    return "_".join([camel_to_snake(c_) for c_ in sorted(c for c in classes)])


def dump_components(
    library_path: Path, library_components: List[Dict[str, Any]]
) -> None:
    library_path.mkdir(exist_ok=True)
    components_final = {
        json_component_name(component["classes"]): [
            cp
            for cp in library_components
            if json_component_name(component["classes"])
            == json_component_name(cp["classes"])
        ]
        for component in library_components
    }
    for file_name, components_ in components_final.items():
        file = library_path / f"{file_name}.yaml"
        with file.open("w") as file_:
            file_.write(
                yaml.dump(
                    components_, Dumper=BlockStyleDumper, default_flow_style=False
                )
            )



import logging
import tempfile
from typing import Callable, List

import pandas as pd
import plotly.graph_objects as go  # type: ignore
from buildingspy.io.outputfile import Reader  # type: ignore
from docx.document import Document
from docx.enum.text import WD_ALIGN_PARAGRAPH
from docx.shared import Inches
from matplotlib import pyplot as plt
from matplotlib.figure import Figure as pyFigure
from plotly.graph_objects import Figure as plotlyFigure
from plotly.subplots import make_subplots  # type: ignore

from trano.elements import BaseElement, NamedFigure

logger = logging.getLogger(__name__)
FIGURE_COUNT = 1


def plot(data: Reader, figure: NamedFigure, show: bool = True) -> pyFigure:
    fig, ax = plt.subplots(figsize=(10, 6))
    fig.subplots_adjust(right=0.9)
    twin1 = ax.twinx()
    plots = []
    tkw = {"size": 4, "width": 1.5}
    for axis, figure_axis in [(ax, figure.left_axis), (twin1, figure.right_axis)]:
        for line in figure_axis.lines:
            try:
                line_data = pd.DataFrame(data.values(line.key))
            except KeyError:
                logger.warning(f"Key {line.key} not found in data")
                continue
            (p,) = axis.plot(
                line_data.loc[1],
                linestyle=line.line_style,
                linewidth=line.line_width,
                color=line.color,
                label=line.label,
            )
            plots.append(p)
            axis.set_ylabel(figure_axis.label)
            axis.yaxis.label.set_color(p.get_color())
            axis.tick_params(axis="y", colors=p.get_color(), **tkw)

    ax.set_xlabel("Simulation time [-]")
    ax.tick_params(axis="x", **tkw)
    ax.legend(handles=plots)
    plt.xticks(rotation=45)  # Rotate the tick labels
    plt.tight_layout()
    if show:
        plt.show()
    return fig


def plot_plot_ly_many(
    data: Reader, figures: List[NamedFigure], show: bool = False
) -> plotlyFigure:

    fig = make_subplots(specs=[[{"secondary_y": True}]])
    for figure in figures:
        for axis, figure_axis in enumerate([figure.left_axis, figure.right_axis]):
            for line in figure_axis.lines:
                try:
                    line_data = pd.DataFrame(data.values(line.key))
                except KeyError:
                    logger.warning(f"Key {line.key} not found in data")
                    continue

                if line_data.empty:
                    continue

                fig.add_trace(
                    go.Scatter(
                        x=line_data.loc[0],
                        y=line_data.loc[1],
                        mode="lines",
                        name=f"{figure.name} - {line.label}",
                    ),
                    secondary_y=bool(axis),
                )
    if not fig.data:
        return None
    fig.update_layout(
        xaxis_title="Simulation time [-]",
        yaxis_title=figure.left_axis.label,
        yaxis2_title=figure.right_axis.label,
        legend_title="Legend",
        autosize=False,
        width=1000,
        height=600,
        margin={"l": 50, "r": 50, "b": 100, "t": 100, "pad": 4},
        plot_bgcolor="rgba(0,0,0,0)",
        paper_bgcolor="rgba(0,0,0,0)",
    )

    # Show the figure
    if show:
        fig.show()

    return fig


def plot_plot_ly(data: Reader, figure: NamedFigure, show: bool = False) -> plotlyFigure:

    fig = make_subplots(specs=[[{"secondary_y": True}]])

    for axis, figure_axis in enumerate([figure.left_axis, figure.right_axis]):
        for line in figure_axis.lines:
            try:
                line_data = pd.DataFrame(data.values(line.key))
            except KeyError:
                logger.warning(f"Key {line.key} not found in data")
                continue

            if line_data.empty:
                continue

            fig.add_trace(
                go.Scatter(
                    x=line_data.loc[0],
                    y=line_data.loc[1],
                    mode="lines",
                    name=line.label,
                ),
                secondary_y=bool(axis),
            )
    if not fig.data:
        return None
    fig.update_layout(
        xaxis_title="Simulation time [-]",
        yaxis_title=figure.left_axis.label,
        yaxis2_title=figure.right_axis.label,
        legend_title="Legend",
        autosize=False,
        width=1000,
        height=600,
        margin={"l": 50, "r": 50, "b": 100, "t": 100, "pad": 4},
    )

    # Show the figure
    if show:
        fig.show()

    return fig


def plot_element(
    data: Reader,
    element: BaseElement,
    plot_function: Callable[[Reader, NamedFigure], plotlyFigure | pyFigure] = plot,
) -> List[pyFigure]:
    figures = []
    subsystems = ["control", "emissions", "ventilation_inlets", "ventilation_outlets"]
    for figure in element.figures:
        fig = plot_function(data, figure)
        if fig:
            figures.append(fig)
        for subsystem in subsystems:
            if hasattr(element, subsystem) and getattr(element, subsystem) is not None:
                sub_element = getattr(element, subsystem)
                if isinstance(sub_element, list):
                    for sub in sub_element:
                        for figure_ in sub.figures:
                            fig = plot_function(data, figure_)
                            figures.append(fig)
                else:
                    for control_figure in sub_element.figures:
                        fig = plot_function(data, control_figure)
                        figures.append(fig)
    return figures


def add_element_figures(document: Document, data: Reader, element: BaseElement) -> None:
    # TODO: duplicate with the one bellow  !!! to be checked!!!
    figures = plot_element(data, element)
    for figure in figures:
        add_figure(document, figure, size=6)


def add_figure(document: Document, fig: pyFigure, size: int = 6) -> None:
    global FIGURE_COUNT  # noqa: PLW0603
    paragraph = document.add_paragraph()
    run = paragraph.add_run()
    run.add_break()
    with tempfile.NamedTemporaryFile(suffix=".jpg") as f:
        fig.savefig(f.name, dpi=1200)
        # Add a figure
        paragraph = document.add_paragraph()
        run = paragraph.add_run()
        run.add_picture(f.name, width=Inches(size))
        paragraph.alignment = WD_ALIGN_PARAGRAPH.CENTER  # Center the image
        caption = document.add_paragraph(
            f"Figure {FIGURE_COUNT}: test", style="Caption"
        )
        FIGURE_COUNT += 1
        caption.alignment = WD_ALIGN_PARAGRAPH.CENTER
    paragraph = document.add_paragraph()
    run = paragraph.add_run()
    run.add_break()



from typing import List, Optional, TYPE_CHECKING, Type, Union

from trano.elements import Control
from trano.elements.base import BaseElement
from trano.elements.types import BaseVariant, ContainerTypes
from pydantic import model_validator, BaseModel

from trano.exceptions import WrongSystemFlowError
import networkx as nx

if TYPE_CHECKING:
    from trano.topology import Network
    from trano.elements import Space


class System(BaseElement):
    control: Optional[Control] = None

    @model_validator(mode="after")
    def _validator(self) -> "System":
        if self.control:
            self.control.container_type = self.container_type
        return self

    def system_ports_connected(self) -> bool:
        return all(port.connected for port in self.ports)


class Sensor(System): ...


class EmissionVariant(BaseVariant):
    radiator: str = "radiator"
    ideal: str = "ideal"


class SpaceSystem(System):
    linked_space: Optional[str] = None


class SpaceHeatingSystem(SpaceSystem):
    container_type: ContainerTypes = "emission"


class Emission(SpaceHeatingSystem): ...


class Ventilation(SpaceSystem):
    container_type: ContainerTypes = "ventilation"


class BaseWeather(System): ...


class BaseOccupancy(System):
    space_name: Optional[str] = None


class DistributionSystem(System):
    container_type: ContainerTypes = "distribution"


class Weather(BaseWeather): ...


class Valve(SpaceHeatingSystem): ...


class ThreeWayValve(DistributionSystem): ...


class TemperatureSensor(Sensor): ...


class SplitValve(DistributionSystem): ...


class Radiator(Emission): ...


class HydronicSystemControl(BaseModel):
    def configure(self, network: "Network") -> None:
        from trano.elements import CollectorControl

        if hasattr(self, "control") and isinstance(self.control, CollectorControl):
            self.control.valves = self._get_linked_valves(network)

    def _get_linked_valves(self, network: "Network") -> List[Valve]:
        valves_: List[Valve] = []
        valves = [node for node in network.graph.nodes if isinstance(node, Valve)]
        for valve in valves:
            paths = list(nx.all_simple_paths(network.graph, self, valve))
            for path in paths:
                p = path[1:-1]
                if p and all(isinstance(p_, System) for p_ in p):
                    valves_.append(valve)
                    break
        return valves_


class Pump(HydronicSystemControl, DistributionSystem): ...


class Occupancy(BaseOccupancy): ...


class Duct(Ventilation): ...


class DamperVariant(BaseVariant):
    complex: str = "complex"


class Damper(Ventilation): ...


class VAV(Damper):
    variant: str = DamperVariant.default


class ProductionSystem(System):
    container_type: ContainerTypes = "production"


class Boiler(HydronicSystemControl, ProductionSystem): ...


class AirHandlingUnit(Ventilation):
    def configure(self, network: "Network") -> None:
        from trano.elements import AhuControl

        if self.control and isinstance(self.control, AhuControl):
            self.control.spaces = self._get_ahu_space_elements(network)
            self.control.vavs = self._get_ahu_vav_elements(network)

    def _get_ahu_space_elements(self, network: "Network") -> List["Space"]:
        from trano.elements import Space

        return [
            x for x in self._get_ahu_elements(Space, network) if isinstance(x, Space)
        ]

    def _get_ahu_vav_elements(self, network: "Network") -> List[VAV]:
        return [x for x in self._get_ahu_elements(VAV, network) if isinstance(x, VAV)]

    def _get_ahu_elements(
        self, element_type: Type[Union[VAV, "Space"]], network: "Network"
    ) -> List[Union[VAV, "Space"]]:
        elements_: List[Union[VAV, "Space"]] = []
        elements = [
            node for node in network.graph.nodes if isinstance(node, element_type)
        ]
        for element in elements:
            try:
                paths = nx.shortest_path(network.graph, self, element)
            except Exception as e:
                raise WrongSystemFlowError(
                    "Wrong AHU system configuration flow."
                ) from e
            p = paths[1:-1]
            if p and all(isinstance(p_, Ventilation) for p_ in p):
                elements_.append(element)
        return elements_



class WrongSystemFlowError(Exception):
    """Exception raised for wrong system flow."""


class ControllerBusPortError(Exception):
    """Exception raised for controller bus port error."""


class IncompatiblePortsError(Exception): ...


class DockerNotInstalledError(Exception): ...


class DockerClientError(Exception): ...


class ConnectionLimitReached(Exception): ...  # noqa: N818


class NoConnectionFoundError(Exception): ...


class ContainerNotFoundError(Exception): ...


class SystemsNotConnectedError(Exception): ...


class UnknownLibraryError(Exception): ...


class UnknownComponentVariantError(Exception): ...



import logging
import os
from typing import Any, Dict, List, Optional

from linkml.generators.pythongen import PythonGenerator  # type: ignore
from linkml.utils import datautils, validation  # type: ignore
from linkml.utils.datautils import (  # type: ignore
    _get_context,
    _get_format,
    _is_xsv,
    get_loader,
    infer_index_slot,
    infer_root_class,
)
from linkml_runtime.linkml_model import Prefix  # type: ignore
from linkml_runtime.utils import inference_utils  # type: ignore
from linkml_runtime.utils.compile_python import compile_python  # type: ignore
from linkml_runtime.utils.inference_utils import infer_all_slot_values  # type: ignore
from linkml_runtime.utils.schemaview import SchemaView  # type: ignore


def delete_none(_dict: Dict[str, Any]) -> Dict[str, Any]:
    # TODO: this function needs to be reviewed
    """Delete None values recursively from all of the dictionaries"""
    for key, value in list(_dict.items()):
        if isinstance(value, dict):
            delete_none(value)
        elif value is None:
            del _dict[key]
        elif isinstance(value, list):
            for v_i in value:
                if isinstance(v_i, dict):
                    delete_none(v_i)

    return _dict


# TODO: function taken from linkml_runtime.cli.converter
def converter(  # noqa: PLR0915, PLR0912, PLR0913, C901
    input: Optional[str] = None,
    module: Optional[Any] = None,  # noqa: ANN401
    target_class: Optional[str] = None,
    context: Optional[List[Any]] = None,
    output: Optional[str] = None,
    input_format: Optional[str] = None,
    output_format: Optional[str] = None,
    prefix: Optional[str] = None,
    target_class_from_path: Optional[str] = None,
    schema: Optional[str] = None,
    validate: Optional[bool] = None,
    infer: Optional[bool] = None,
    index_slot: Optional[bool] = None,
) -> Dict[str, Any]:
    """
    Converts instance data to and from different LinkML Runtime serialization formats.

    The instance data must conform to a LinkML model, and either a path to a python
    module must be passed, or a path to a schema.

    The converter works by first using a linkml-runtime *loader* to
    instantiate in-memory model objects, then a *dumper* is used to serialize.
    A validation step is optionally performed in between

    When converting to or from RDF, a path to a schema must be provided.

    For more information, see https://linkml.io/linkml/data/index.html
    """
    if prefix is None:
        prefix = []  # type: ignore
    if module is None:
        if schema is None:
            raise Exception("must pass one of module OR schema")
        else:
            python_module = PythonGenerator(schema).compile_module()
    else:
        python_module = compile_python(module)
    prefix_map = {}
    if prefix:
        for p in prefix:
            base, uri = p.split("=")
            prefix_map[base] = uri
    if schema is not None:
        sv = SchemaView(schema)
        if prefix_map:
            for k, v in prefix_map.items():
                sv.schema.prefixes[k] = Prefix(k, v)
                sv.set_modified()
    if target_class is None and target_class_from_path:
        target_class = os.path.basename(input).split("-")[0]  # type: ignore
        logging.info(f"inferred target class = {target_class} from {input}")
    if target_class is None:
        target_class = infer_root_class(sv)
    if target_class is None:
        raise Exception("target class not specified and could not be inferred")
    py_target_class = python_module.__dict__[target_class]
    input_format = _get_format(input, input_format)
    loader = get_loader(input_format)

    inargs = {}
    outargs = {}
    if datautils._is_rdf_format(input_format):
        if sv is None:
            raise Exception("Must pass schema arg")
        inargs["schemaview"] = sv
        inargs["fmt"] = input_format
    if _is_xsv(input_format):
        if index_slot is None:
            index_slot = infer_index_slot(sv, target_class)
            if index_slot is None:
                raise Exception("--index-slot is required for CSV input")
        inargs["index_slot"] = index_slot
        inargs["schema"] = schema
    obj = loader.load(source=input, target_class=py_target_class, **inargs)
    if infer:
        infer_config = inference_utils.Config(
            use_expressions=True, use_string_serialization=True
        )
        infer_all_slot_values(obj, schemaview=sv, config=infer_config)
    if validate:
        if schema is None:
            raise Exception(
                "--schema must be passed in order to validate. Suppress with --no-validate"
            )
        # TODO: use validator framework
        validation.validate_object(obj, schema)

    output_format = _get_format(output, output_format, default="json")
    if output_format == "json-ld":
        if len(context) == 0:  # type: ignore
            if schema is not None:
                context = [_get_context(schema)]
            else:
                raise Exception("Must pass in context OR schema for RDF output")
        outargs["contexts"] = list(context)  # type: ignore
    if output_format in ["rdf", "ttl"]:
        if sv is None:
            raise Exception("Must pass schema arg")
        outargs["schemaview"] = sv
    if _is_xsv(output_format):
        if index_slot is None:
            index_slot = infer_index_slot(sv, target_class)
            if index_slot is None:
                raise Exception("--index-slot is required for CSV output")
        outargs["index_slot"] = index_slot  # type: ignore
        outargs["schema"] = schema  # type: ignore
    converted_data = delete_none(obj._as_dict)
    return converted_data



from typing import Optional, Tuple

from pydantic import BaseModel, Field, ConfigDict

from trano.elements.types import ContainerTypes


class BaseProperties(BaseModel):
    container_type: ContainerTypes
    is_package: bool
    data: str


class MandatoryPoint(BaseModel):
    x: float
    y: float


class Point(BaseModel):
    x: Optional[float] = None
    y: Optional[float] = None

    def is_valid(self) -> bool:
        return self.x is not None and self.y is not None

    @property
    def c_(self) -> MandatoryPoint:
        return MandatoryPoint(x=self.x, y=self.y)


class ElementPosition(BaseModel):
    location: Point = Field(default_factory=Point)
    annotation: str = """annotation (
    Placement(transformation(origin = {{ macros.join_list(element.position.global_.coordinate()) }},
    extent = {% raw %}{{10, -10}, {-10, 10}}
    {% endraw %})));"""

    def is_empty(self) -> bool:
        return not self.location.is_valid()

    def coordinate(self) -> Tuple[float, float]:
        return self.location.c_.x, self.location.c_.y


class BasePosition(BaseModel):
    container: ElementPosition = Field(
        default_factory=lambda: ElementPosition(
            annotation="""annotation (
    Placement(transformation(origin = {{ macros.join_list(element.position.container.coordinate()) }},
    extent = {% raw %}{{10, -10}, {-10, 10}}
    {% endraw %})));"""
        )
    )
    global_: ElementPosition = Field(default_factory=ElementPosition)

    def set_global(self, x: float, y: float) -> None:
        self.global_.location.x = x
        self.global_.location.y = y

    def set_container(self, x: float, y: float) -> None:
        self.container.location.x = x
        self.container.location.y = y

    @property
    def x_container(self) -> float:
        return self.container.location.c_.x

    @property
    def y_container(self) -> float:
        return self.container.location.c_.y

    def is_global_empty(self) -> bool:
        return self.global_.is_empty()

    def is_container_empty(self) -> bool:
        return self.container.is_empty()

    def set(self, x: float, y: float) -> None:
        self.set_global(x, y)
        self.set_container(x, y)

    def between_two_objects(
        self, position_1: ElementPosition, position_2: ElementPosition
    ) -> None:
        self.set(
            (position_1.location.c_.x - position_2.location.c_.x) / 2,
            (position_1.location.c_.y + position_2.location.c_.y) / 2,
        )


class ComponentModel(BaseModel):
    id: int
    model: str
    container: str


class BaseElementPosition(BaseModel):
    position: BasePosition = Field(default_factory=BasePosition)
    model_config = ConfigDict(validate_assignment=True)


class BaseParameter(BaseModel):
    model_config = ConfigDict(populate_by_name=True, extra="allow")


class MediumTemplate(BaseModel):
    air: Optional[str] = None
    water: Optional[str] = None

    def is_empty(self) -> bool:
        return self.air is None or self.water is None



import json
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, cast

from pydantic import BaseModel, Field

from trano.elements.common_base import MediumTemplate
from trano.exceptions import UnknownLibraryError

if TYPE_CHECKING:
    from trano.elements import WallParameters


# TODO: this must go!!!
def tilts_processing_ideas(element: "WallParameters") -> List[str]:
    return [f"IDEAS.Types.Tilt.{tilt.value.capitalize()}" for tilt in element.tilts]


class Templates(BaseModel):
    is_package: bool = False
    construction: str
    glazing: str
    material: Optional[str] = None
    main: str


def read_libraries() -> Dict[str, Dict[str, Any]]:
    library_json_path = Path(__file__).parent.joinpath("library.json")
    return cast(Dict[str, Dict[str, Any]], json.loads(library_json_path.read_text()))


class Library(BaseModel):
    name: str
    merged_external_boundaries: bool = False
    functions: Dict[str, Callable[[Any], Any]] = {
        "tilts_processing_ideas": tilts_processing_ideas
    }
    core_library: Optional[str] = None
    medium: MediumTemplate
    constants: str = ""
    templates: Templates
    default: bool = False
    default_parameters: Dict[str, Any] = Field(
        default_factory=dict
    )  # TODO: this should be baseparameters

    def base_library(self) -> str:
        return self.core_library or self.name

    @classmethod
    def from_configuration(cls, name: str) -> "Library":
        libraries = read_libraries()

        if name not in libraries:
            raise UnknownLibraryError(
                f"Library {name} not found. Available libraries: {list(libraries)}"
            )
        library_data = libraries[name]
        return cls(**library_data)

    @classmethod
    def load_default(cls) -> "Library":
        libraries = read_libraries()
        default_library = [
            library_data
            for _, library_data in libraries.items()
            if library_data.get("default")
        ]
        if not default_library:
            raise ValueError("No default library found")
        return cls(**default_library[0])



from typing import List, Optional, Literal

from pydantic import BaseModel, computed_field, field_validator


class Target(BaseModel):
    main: str
    sub: Optional[str] = None
    evaluated_element: str = ""
    value: str = ""

    @field_validator("value")
    @classmethod
    def validate_value(cls, value: str) -> str:
        return value.capitalize()

    def __hash__(self) -> int:
        return hash((self.main, self.sub))

    def commands(self) -> list[str]:
        return self.main.split(".")[1:]

    def sub_commands(self) -> List[str]:
        if self.sub is None:
            raise Exception("Target sub is None")
        return self.sub.split(".")


LoadType = Literal["Resistive", "Inductive", "Capacitive"]
PrimaryEnergyType = Literal["Electricity", "Gas"]


class PowerInput(BaseModel):
    energy: PrimaryEnergyType
    load_type: Optional[LoadType] = None

    def name(self) -> str:
        if self.load_type:
            return f"{self.energy.lower()}_{self.load_type}"
        return self.energy.lower()


class BaseInputOutput(BaseModel):
    name: str
    component: str
    port: str
    multi: bool = False
    target: Target
    input_template: str
    default: float | str | int
    power: Optional[PowerInput] = None

    def __hash__(self) -> int:
        return hash((self.name, self.target.value))

    def __eq__(self, other: object) -> bool:
        return hash(self) == hash(other)

    @computed_field  # type: ignore
    @property
    def input_model(self) -> str:
        if self.target.evaluated_element:
            return f"""{self.input_template}
            {self.name}{self.target.evaluated_element.capitalize()}
            (y={self.default});"""
        return ""

    @computed_field  # type: ignore
    @property
    def input_name(self) -> str:
        if self.power is None:
            return self.name
        return self.power.name()


class BaseInput(BaseInputOutput): ...


class RealInput(BaseInput):
    default: float = 0.0
    input_template: str = "Modelica.Blocks.Sources.RealExpression"


class IntegerInput(BaseInput):
    default: int = 0
    input_template: str = "Modelica.Blocks.Sources.IntegerExpression"


class BooleanInput(BaseInput):
    default: str = "false"
    input_template: str = "Modelica.Blocks.Sources.BooleanExpression"


class BooleanOutput(BaseInputOutput):
    default: str = "false"
    input_template: str = "Modelica.Blocks.Sources.BooleanExpression"


class IntegerOutput(BaseInputOutput):
    default: int = 0
    input_template: str = "Modelica.Blocks.Sources.IntegerExpression"


class RealOutput(BaseInputOutput):
    default: float = 0.0
    input_template: str = "Modelica.Blocks.Sources.RealExpression"
